"""Tests for the Mailgun SMTP adapter."""

import typing as t
from collections.abc import AsyncGenerator, Callable
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from httpx import Response as HttpxResponse
from pydantic import SecretStr
from acb.adapters.dns._base import DnsRecord
from acb.adapters.smtp.mailgun import Smtp as MailgunSmtp
from acb.adapters.smtp.mailgun import SmtpSettings as MailgunSmtpSettings


@pytest.fixture
def mock_async_context_manager() -> Callable[
    ..., AbstractAsyncContextManager[MagicMock]
]:
    @asynccontextmanager
    async def _async_context_manager(
        *args: Any, **kwargs: Any
    ) -> AsyncGenerator[MagicMock]:
        yield MagicMock()

    return _async_context_manager


class TestMailgunSmtpSettings:
    def test_init_basic(self) -> None:
        settings: MailgunSmtpSettings = MailgunSmtpSettings(
            name="mailgun",
            domain="example.com",
            from_email="test@example.com",
            api_key="test_api_key",
            route_name="route_name",
            priority=10,
            actions=["forward('example.com')"],
            description="Test Mailgun SMTP",
            expression="match_recipient('.*@example.com')",
        )
        assert settings.name == "mailgun"  # type: ignore
        assert settings.domain == "example.com"
        assert settings.from_email == "test@example.com"  # type: ignore
        assert settings.api_key == "test_api_key"
        assert settings.route_name == "route_name"  # type: ignore
        assert settings.priority == 10  # type: ignore
        assert settings.actions == ["forward('example.com')"]  # type: ignore
        assert settings.description == "Test Mailgun SMTP"  # type: ignore
        assert settings.expression == "match_recipient('.*@example.com')"  # type: ignore

    def test_init_with_config(self) -> None:
        mock_config: MagicMock = MagicMock()
        mock_config.app.domain = "example.com"
        mock_config.app.title = "Example App"
        mock_config.smtp = MagicMock()
        mock_config.smtp.api_key = SecretStr("test-api-key")
        mock_config.smtp.password = SecretStr("test-password")

        test_values = {
            "name": "mailgun",
            "domain": "mail.example.com",
            "from_email": "info@example.com",
            "api_key": SecretStr("test-api-key"),
        }

        with patch("acb.depends.depends.__call__", return_value=mock_config):
            settings = MailgunSmtpSettings(**test_values)

            assert settings.domain == "mail.example.com"
            assert settings.from_email == "info@example.com"  # type: ignore
            assert settings.api_url == "https://api.mailgun.net/v3/domains"
            assert settings.mx_servers == ["smtp.mailgun.com"]
            assert settings.api_key.get_secret_value() == "test-api-key"
            assert settings.password.get_secret_value() == "test-password"


class TestMailgunSmtp:
    @pytest.fixture
    def smtp(
        self,
        mock_async_context_manager: Callable[
            ..., AbstractAsyncContextManager[MagicMock]
        ],
    ) -> MailgunSmtp:
        class TestableMailgunSmtp(MailgunSmtp):
            def __init__(self) -> None:
                super().__init__()
                self.connect = MagicMock()

        instance: MailgunSmtp = TestableMailgunSmtp()
        instance.config = MagicMock()
        instance.config.smtp = MagicMock()
        instance.config.smtp.api_url = "https://api.mailgun.net/v3/domains"
        instance.config.smtp.api_key = MagicMock()
        instance.config.smtp.api_key.get_secret_value.return_value = "test-api-key"
        instance.config.smtp.domain = "example.com"
        instance.config.app = MagicMock()
        instance.config.app.domain = "example.com"
        instance.config.smtp.forwards = {"admin": "admin@example.com"}
        instance.logger = MagicMock()
        instance.requests = MagicMock()

        if not hasattr(instance, "get_response") or instance.get_response is None:
            instance.get_response = AsyncMock(return_value={"message": "success"})

        return instance

    @pytest.mark.asyncio
    async def test_get_response_get(self, smtp: MailgunSmtp) -> None:
        async def mock_get_response_wrapper(
            req_type: str,
            domain: str | None = None,
            data: dict[str, t.Any] | None = None,
            params: dict[str, int] | None = None,
        ) -> dict[str, t.Any]:
            if req_type == "get":
                mock_response = MagicMock(spec=HttpxResponse)
                mock_response.json.return_value = {"message": "success"}
                smtp.requests.get = AsyncMock(return_value=mock_response)
                await smtp.requests.get()
                return {"message": "success"}
            return {}

        smtp.get_response = mock_get_response_wrapper

        with patch(
            "acb.adapters.smtp.mailgun.load.json",
            AsyncMock(return_value={"message": "success"}),
        ):
            result = await smtp.get_response(
                "get", domain="example.com", params={"limit": 10}
            )

            smtp.requests.get.assert_called_once()
            assert result == {"message": "success"}

    @pytest.mark.asyncio
    async def test_get_response_put(self, smtp: MailgunSmtp) -> None:
        async def mock_get_response_wrapper(
            req_type: str,
            domain: str | None = None,
            data: dict[str, t.Any] | None = None,
            params: dict[str, int] | None = None,
        ) -> dict[str, t.Any]:
            if req_type == "put":
                mock_response = MagicMock(spec=HttpxResponse)
                mock_response.json.return_value = {"message": "success"}
                smtp.requests.put = AsyncMock(return_value=mock_response)
                await smtp.requests.put()
                return {"message": "success"}
            return {}

        smtp.get_response = mock_get_response_wrapper

        with (
            patch(
                "acb.adapters.smtp.mailgun.load.json",
                AsyncMock(return_value={"message": "success"}),
            ),
            patch("acb.adapters.smtp.mailgun.sys._getframe") as mock_frame,
        ):
            mock_frame.return_value.f_back.f_code.co_name = "test_domain"

            result = await smtp.get_response(
                "put", domain="example.com", data={"key": "value"}
            )

            smtp.requests.put.assert_called_once()
            assert result == {"message": "success"}

    @pytest.mark.asyncio
    async def test_list_domains(self, smtp: MailgunSmtp) -> None:
        with patch.object(smtp, "get_response", AsyncMock()) as mock_get_response:
            mock_get_response.return_value = {
                "items": [{"name": "example.com"}, {"name": "test.com"}]
            }

            domains = await smtp.list_domains()

            mock_get_response.assert_called_once_with(
                "get", params={"skip": 0, "limit": 1000}
            )
            assert domains == ["example.com", "test.com"]

    @pytest.mark.asyncio
    async def test_get_domain(self, smtp: MailgunSmtp) -> None:
        with patch.object(smtp, "get_response", AsyncMock()) as mock_get_response:
            mock_get_response.return_value = {
                "domain": "example.com",
                "state": "active",
            }

            domain_info = await smtp.get_domain("example.com")

            mock_get_response.assert_called_once_with("get", domain="example.com")
            assert domain_info == {"domain": "example.com", "state": "active"}

    @pytest.mark.asyncio
    async def test_create_domain(self, smtp: MailgunSmtp) -> None:
        with patch.object(smtp, "get_response", AsyncMock()) as mock_get_response:
            mock_get_response.side_effect = [
                {"message": "Domain created"},
                {"message": "Connection updated"},
            ]

            result = await smtp.create_domain("example.com")

            assert mock_get_response.call_count == 2
            assert result == {"message": "Connection updated"}

    @pytest.mark.asyncio
    async def test_delete_domain(self, smtp: MailgunSmtp) -> None:
        with patch.object(smtp, "get_response", AsyncMock()) as mock_get_response:
            mock_get_response.return_value = {"message": "Domain deleted"}

            result = await smtp.delete_domain("example.com")

            mock_get_response.assert_called_once_with("delete", domain="example.com")
            assert result == {"message": "Domain deleted"}

    @pytest.mark.asyncio
    async def test_get_dns_records(self, smtp: MailgunSmtp) -> None:
        with (
            patch.object(smtp, "get_domain", AsyncMock()) as mock_get_domain,
            patch("acb.adapters.smtp.mailgun.debug") as mock_debug,
        ):
            mock_get_domain.return_value = {
                "receiving_dns_records": [
                    {"priority": "10", "value": "mx.example.com"}
                ],
                "sending_dns_records": [
                    {"name": "example.com", "record_type": "TXT", "value": "v=spf1"}
                ],
            }

            records = await smtp.get_dns_records("example.com")

            mock_get_domain.assert_called_once_with("example.com")
            assert len(records) == 2
            assert records[0].type == "MX"
            assert records[1].type == "TXT"
            mock_debug.assert_called_once()

    @pytest.mark.asyncio
    async def test_create_dns_records(self, smtp: MailgunSmtp) -> None:
        mock_dns = AsyncMock()

        with (
            patch.object(smtp, "create_domain", AsyncMock()) as mock_create_domain,
            patch.object(
                smtp, "create_domain_credentials", AsyncMock()
            ) as mock_create_creds,
            patch.object(smtp, "get_dns_records", AsyncMock()) as mock_get_records,
            patch.object(smtp, "delete_domain", AsyncMock()) as mock_delete_domain,
        ):
            mock_records = [MagicMock(spec=DnsRecord), MagicMock(spec=DnsRecord)]
            mock_get_records.return_value = mock_records
            smtp.config.mail.mailgun.gmail.enabled = False

            await smtp.create_dns_records(dns=mock_dns)

            mock_create_domain.assert_called_once_with(smtp.config.smtp.domain)
            mock_create_creds.assert_called_once_with(smtp.config.smtp.domain)
            mock_get_records.assert_called_once_with(smtp.config.smtp.domain)
            mock_delete_domain.assert_called_once_with(smtp.config.smtp.domain)
            mock_dns.create_records.assert_called_once_with(mock_records)

    @pytest.mark.asyncio
    async def test_create_route(self, smtp: MailgunSmtp) -> None:
        with (
            patch.object(smtp, "list_routes", AsyncMock()) as mock_list_routes,
            patch.object(smtp, "get_response", AsyncMock()) as mock_get_response,
        ):
            mock_list_routes.return_value = []
            mock_get_response.return_value = {"message": "Route created"}

            # Create mock logger
            mock_logger = MagicMock()
            smtp.logger = mock_logger

            result = await smtp.create_route("admin", "admin@example.com")

            mock_list_routes.assert_called_once()
            mock_get_response.assert_called_once()
            assert result == {"message": "Route created"}
            mock_logger.info.assert_called_once()

    @pytest.mark.asyncio
    async def test_init(self, smtp: MailgunSmtp) -> None:
        with (
            patch.object(smtp, "create_dns_records", AsyncMock()) as mock_create_dns,
            patch.object(smtp, "delete_routes", AsyncMock()) as mock_delete_routes,
            patch.object(smtp, "create_routes", AsyncMock()) as mock_create_routes,
        ):
            await smtp.init()

            mock_create_dns.assert_called_once()
            mock_delete_routes.assert_called_once()
            mock_create_routes.assert_called_once()



"""Tests for the FTP adapter."""

import tempfile
import typing as t
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from aioftp import Client as AioFtpClient
from acb.adapters.ftpd._base import FileInfo
from acb.adapters.ftpd.ftp import Ftpd, FtpdSettings


@t.runtime_checkable
class MockFtpClientProtocol(t.Protocol):
    _connect: AsyncMock
    _login: AsyncMock
    _quit: AsyncMock
    _upload: AsyncMock
    _download: AsyncMock
    _list: AsyncMock
    _make_directory: AsyncMock
    _remove_directory: AsyncMock
    _remove_file: AsyncMock
    _rename: AsyncMock
    _stat: AsyncMock

    async def connect(self, host: str, port: int) -> None: ...
    async def login(self, username: str, password: str) -> None: ...
    async def quit(self) -> None: ...
    async def upload(self, local_path: str, remote_path: str) -> None: ...
    async def download(self, remote_path: str, local_path: str) -> None: ...
    def list(self, path: str) -> t.AsyncIterator[tuple[None, dict[str, t.Any]]]: ...
    async def make_directory(self, path: str) -> None: ...
    async def remove_directory(self, path: str) -> None: ...
    async def remove_file(self, path: str) -> None: ...
    async def rename(self, old_path: str, new_path: str) -> None: ...
    async def stat(self, path: str) -> dict[str, t.Any]: ...


class MockFtpClient(MockFtpClientProtocol):
    def __init__(self) -> None:
        self._connect = AsyncMock()
        self._login = AsyncMock()
        self._quit = AsyncMock()
        self._upload = AsyncMock()
        self._download = AsyncMock()
        self._list = AsyncMock()
        self._make_directory = AsyncMock()
        self._remove_directory = AsyncMock()
        self._remove_file = AsyncMock()
        self._rename = AsyncMock()
        self._stat = AsyncMock()

    async def connect(self, host: str, port: int) -> None:
        return await self._connect(host, port)

    async def login(self, username: str, password: str) -> None:
        return await self._login(username, password)

    async def quit(self) -> None:
        return await self._quit()

    async def upload(self, local_path: str, remote_path: str) -> None:
        return await self._upload(local_path, remote_path)

    async def download(self, remote_path: str, local_path: str) -> None:
        return await self._download(remote_path, local_path)

    def list(self, path: str) -> t.AsyncIterator[tuple[None, dict[str, t.Any]]]:
        async def _list_generator() -> t.AsyncIterator[tuple[None, dict[str, t.Any]]]:
            file_infos = await self._list(path)
            for info in file_infos:
                yield info

        return _list_generator()

    async def make_directory(self, path: str) -> None:
        return await self._make_directory(path)

    async def remove_directory(self, path: str) -> None:
        return await self._remove_directory(path)

    async def remove_file(self, path: str) -> None:
        return await self._remove_file(path)

    async def rename(self, old_path: str, new_path: str) -> None:
        return await self._rename(old_path, new_path)

    async def stat(self, path: str) -> dict[str, t.Any]:
        return await self._stat(path)


class MockServer:
    def __init__(self, **kwargs: t.Any) -> None:
        self.kwargs = kwargs
        self._start = AsyncMock()
        self._close = AsyncMock()

    async def start(self) -> None:
        return await self._start()

    async def close(self) -> None:
        return await self._close()


class MockPathIO:
    def __init__(self) -> None:
        pass


@pytest.fixture
def mock_server() -> MockServer:
    return MockServer()


@pytest.fixture
def mock_client() -> MockFtpClient:
    client = MockFtpClient()

    client._connect.return_value = None
    client._login.return_value = None
    client._quit.return_value = None
    client._upload.return_value = None
    client._download.return_value = None
    client._make_directory.return_value = None
    client._remove_directory.return_value = None
    client._remove_file.return_value = None
    client._rename.return_value = None

    file_infos = [
        (
            None,
            {
                "name": "file1.txt",
                "size": 100,
                "type": "file",
                "permissions": "rw-r--r--",
                "modify": 1621234567.0,
                "owner": "user",
                "group": "users",
            },
        ),
        (
            None,
            {
                "name": "folder1",
                "size": 0,
                "type": "dir",
                "permissions": "rwxr-xr-x",
                "modify": 1621234567.0,
                "owner": "user",
                "group": "users",
            },
        ),
    ]
    client._list.return_value = file_infos

    client._stat.return_value = {
        "name": "file.txt",
        "size": 100,
        "type": "file",
        "permissions": "rw-r--r--",
        "modify": 1621234567.0,
        "owner": "user",
        "group": "users",
    }

    return client


@pytest.fixture
async def ftpd_adapter(
    mock_client: MockFtpClient, mock_server: MockServer
) -> t.AsyncGenerator[Ftpd]:
    with (
        patch("aioftp.Client", return_value=mock_client),
        patch("aioftp.Server", return_value=mock_server),
    ):
        mock_config = MagicMock()
        mock_config.ftpd.root_directory = tempfile.gettempdir()
        mock_config.ftpd.host = "127.0.0.1"
        mock_config.ftpd.port = 8021
        mock_config.ftpd.username = "user"
        mock_config.ftpd.password = MagicMock()
        mock_config.ftpd.password.get_secret_value.return_value = "password"
        mock_config.ftpd.passive_ports_min = 50000
        mock_config.ftpd.passive_ports_max = 50100
        mock_config.ftpd.timeout = 30
        mock_config.ftpd.anonymous = False
        mock_config.ftpd.root_dir = tempfile.gettempdir()

        adapter = Ftpd()
        adapter.config = mock_config
        adapter._client = mock_client

        adapter.__class__.server = property(lambda self: mock_server)

        yield adapter


class TestFtpdSettings:
    def test_default_values(self) -> None:
        settings = FtpdSettings()

        assert settings.port == 8021
        assert settings.passive_ports_min == 50000
        assert settings.passive_ports_max == 50100
        assert settings.timeout == 30


@pytest.mark.asyncio
async def test_start_server(ftpd_adapter: Ftpd, mock_server: MockServer) -> None:
    mock_logger = MagicMock()

    await ftpd_adapter.start(logger=mock_logger)

    mock_server._start.assert_called_once()
    mock_logger.info.assert_called_once()

    mock_server._start.side_effect = Exception("Server start error")

    with pytest.raises(Exception, match="Server start error"):
        await ftpd_adapter.start(logger=mock_logger)

    mock_server._close.assert_called_once()
    mock_logger.error.assert_called_once()


@pytest.mark.asyncio
async def test_stop_server(ftpd_adapter: Ftpd, mock_server: MockServer) -> None:
    mock_logger = MagicMock()

    await ftpd_adapter.stop(logger=mock_logger)

    mock_server._close.assert_called_once()
    mock_logger.info.assert_called_once_with("FTP server stopped")

    mock_server._close.side_effect = Exception("Server stop error")
    mock_logger.reset_mock()

    with pytest.raises(Exception, match="Server stop error"):
        await ftpd_adapter.stop(logger=mock_logger)

    mock_logger.error.assert_called_once()


@pytest.mark.asyncio
async def test_ensure_client(ftpd_adapter: Ftpd, mock_client: MockFtpClient) -> None:
    ftpd_adapter._client = None

    async def mock_ensure_client() -> AioFtpClient:
        ftpd_adapter._client = t.cast(AioFtpClient, mock_client)
        return t.cast(AioFtpClient, mock_client)

    original_ensure_client = ftpd_adapter._ensure_client

    try:
        ftpd_adapter._ensure_client = mock_ensure_client

        path = "/test_path"
        await ftpd_adapter.list_dir(path)

        mock_client._list.assert_called_once_with(path)
    finally:
        ftpd_adapter._ensure_client = original_ensure_client


@pytest.mark.asyncio
async def test_upload(ftpd_adapter: Ftpd, mock_client: MockFtpClient) -> None:
    temp_dir = tempfile.gettempdir()
    local_path = Path(temp_dir) / "local_file.txt"
    remote_path = "/remote/file.txt"

    await ftpd_adapter.upload(local_path, remote_path)

    mock_client._upload.assert_called_once_with(str(local_path), remote_path)


@pytest.mark.asyncio
async def test_download(ftpd_adapter: Ftpd, mock_client: MockFtpClient) -> None:
    remote_path = "/remote/file.txt"
    temp_dir = tempfile.gettempdir()
    local_path = Path(temp_dir) / "local_file.txt"

    await ftpd_adapter.download(remote_path, local_path)

    mock_client._download.assert_called_once_with(remote_path, str(local_path))


@pytest.mark.asyncio
async def test_list_dir(ftpd_adapter: Ftpd, mock_client: MockFtpClient) -> None:
    path = "/remote"

    result = await ftpd_adapter.list_dir(path)

    mock_client._list.assert_called_once_with(path)
    assert len(result) == 2
    assert isinstance(result[0], FileInfo)
    assert result[0].name == "file1.txt"
    assert result[1].name == "folder1"


@pytest.mark.asyncio
async def test_mkdir(ftpd_adapter: Ftpd, mock_client: MockFtpClient) -> None:
    path = "/remote/new_dir"

    await ftpd_adapter.mkdir(path)

    mock_client._make_directory.assert_called_once_with(path)


@pytest.mark.asyncio
async def test_rmdir_non_recursive(
    ftpd_adapter: Ftpd, mock_client: MockFtpClient
) -> None:
    path = "/remote/empty_dir"

    await ftpd_adapter.rmdir(path)

    mock_client._remove_directory.assert_called_once_with(path)
    mock_client._list.assert_not_called()


@pytest.mark.asyncio
async def test_rmdir_recursive(ftpd_adapter: Ftpd, mock_client: MockFtpClient) -> None:
    original_list_dir = ftpd_adapter.list_dir
    original_delete = ftpd_adapter.delete

    try:
        call_tracker = {"list_dir": [], "delete": [], "rmdir": []}

        async def mock_list_dir(path: str) -> list[FileInfo]:
            call_tracker["list_dir"].append(path)
            if path == "/test/dir":
                return [
                    FileInfo(
                        name="file1.txt",
                        size=100,
                        permissions="rw-r--r--",
                        mtime=1621234567.0,
                    ),
                    FileInfo(
                        name="subdir",
                        is_file=False,
                        is_dir=True,
                        permissions="rwxr-xr-x",
                        mtime=1621234567.0,
                    ),
                ]
            return []

        async def mock_delete(path: str) -> None:
            call_tracker["delete"].append(path)
            return None

        async def mock_rmdir_non_recursive(path: str) -> None:
            call_tracker["rmdir"].append(path)
            return None

        ftpd_adapter.list_dir = mock_list_dir
        ftpd_adapter.delete = mock_delete

        async def mock_rmdir(path: str, recursive: bool = False) -> None:
            if recursive:
                files = await ftpd_adapter.list_dir(path)
                for file in files:
                    full_path = f"{path}/{file.name}"
                    if file.is_dir:
                        await mock_rmdir(full_path, recursive=True)
                    else:
                        await ftpd_adapter.delete(full_path)
                await mock_rmdir_non_recursive(path)
            else:
                await mock_rmdir_non_recursive(path)

        ftpd_adapter.rmdir = mock_rmdir

        path = "/test/dir"
        await ftpd_adapter.rmdir(path, recursive=True)

        assert call_tracker["list_dir"] == ["/test/dir", "/test/dir/subdir"]
        assert call_tracker["delete"] == ["/test/dir/file1.txt"]
        assert call_tracker["rmdir"] == ["/test/dir/subdir", "/test/dir"]
    finally:
        ftpd_adapter.list_dir = original_list_dir
        ftpd_adapter.delete = original_delete


@pytest.mark.asyncio
async def test_delete(ftpd_adapter: Ftpd, mock_client: MockFtpClient) -> None:
    path = "/remote/file.txt"

    await ftpd_adapter.delete(path)

    mock_client._remove_file.assert_called_once_with(path)


@pytest.mark.asyncio
async def test_rename(ftpd_adapter: Ftpd, mock_client: MockFtpClient) -> None:
    old_path = "/remote/old.txt"
    new_path = "/remote/new.txt"

    await ftpd_adapter.rename(old_path, new_path)

    mock_client._rename.assert_called_once_with(old_path, new_path)


@pytest.mark.asyncio
async def test_exists_true(ftpd_adapter: Ftpd, mock_client: MockFtpClient) -> None:
    path = "/remote/file.txt"

    result = await ftpd_adapter.exists(path)

    mock_client._stat.assert_called_once_with(path)
    assert result


@pytest.mark.asyncio
async def test_exists_false(ftpd_adapter: Ftpd, mock_client: MockFtpClient) -> None:
    path = "/remote/nonexistent.txt"
    mock_client._stat.side_effect = Exception("File not found")

    result = await ftpd_adapter.exists(path)

    mock_client._stat.assert_called_once_with(path)
    assert not result


@pytest.mark.asyncio
async def test_stat(ftpd_adapter: Ftpd, mock_client: MockFtpClient) -> None:
    path = "/remote/file.txt"

    result = await ftpd_adapter.stat(path)

    mock_client._stat.assert_called_once_with(path)
    assert isinstance(result, FileInfo)
    assert result.name == "file.txt"
    assert result.size == 100
    assert result.is_file
    assert not result.is_dir


@pytest.mark.asyncio
async def test_read_text(ftpd_adapter: Ftpd, mock_client: MockFtpClient) -> None:
    with (
        patch("tempfile.mkdtemp") as mock_mkdtemp,
        patch("acb.adapters.ftpd.ftp.Path") as mock_path,
        patch("acb.adapters.ftpd.ftp.AsyncPath") as mock_async_path,
        patch("acb.adapters.ftpd.ftp.os"),
    ):
        temp_dir = tempfile.gettempdir()
        safe_temp_dir = f"{temp_dir}/temp_dir"
        mock_mkdtemp_typed: t.Any = mock_mkdtemp
        mock_mkdtemp_typed.return_value = safe_temp_dir

        mock_temp_path = MagicMock()
        mock_path_typed: t.Any = mock_path
        mock_path_typed.return_value = mock_temp_path

        truediv_method: t.Any = mock_temp_path.__truediv__
        truediv_method.return_value = mock_temp_path

        str_method: t.Any = mock_temp_path.__str__
        str_method.return_value = f"{safe_temp_dir}/file.txt"

        mock_temp_path.exists = MagicMock(return_value=True)
        mock_temp_path.unlink = MagicMock()

        # Mock AsyncPath and its read_text method
        mock_async_path_instance = MagicMock()
        mock_async_path.return_value = mock_async_path_instance
        mock_async_path_instance.read_text = AsyncMock(return_value="file contents")

        path = "/remote/file.txt"
        result = await ftpd_adapter.read_text(path)

        assert result == "file contents"
        mock_async_path_instance.read_text.assert_called_once()
        mock_client._download.assert_called_once()



import typing as t
from calendar import isleap, monthrange
from datetime import date, datetime, timedelta
from itertools import chain
from pathlib import Path
from re import search

import arrow
from anyio import Path as AsyncPath
from pydantic import BaseModel, ConfigDict, create_model
from sqlmodel import SQLModel, select
from acb.actions.encode import load
from acb.adapters import import_adapter
from acb.config import Config
from acb.depends import depends
from acb.logger import Logger

Sql, Storage = import_adapter()


class SqlBackupUtils(BaseModel):
    storage: Storage = depends()
    config: Config = depends()
    backup_path: Path | None = None

    @staticmethod
    def get_timestamp(name: str) -> str | None:
        pattern = "(.*)(-)(?P<timestamp>[\\d]{10})(-)(.*)"
        match = search(pattern, name)
        return match.group("timestamp") or None

    def get_files(self) -> t.Generator[t.Any, t.Any, t.Any]:
        blobs = self.storage.sql.list()
        if blobs is not None:
            for blob in blobs:
                if self.get_timestamp(blob.name):
                    yield blob.name

    def get_timestamps(self) -> list[int]:
        if not self.files:
            self.files = tuple(self.get_files())
        different_timestamps = []
        for name in self.files:
            timestamp = self.get_timestamp(name)
            if timestamp and timestamp not in different_timestamps:
                different_timestamps.append(int(timestamp))
        return different_timestamps

    def by_timestamp(self, timestamp: str) -> t.Generator[t.Any, t.Any, t.Any]:
        if not self.files:
            self.files = tuple(self.get_files())
        for name in self.files:
            if timestamp == self.get_timestamp(name):
                yield name

    def valid(self, timestamp: int) -> bool:
        if timestamp in self.get_timestamps():
            return True
        return False

    def get_path(
        self, class_name: str = "", timestamp: t.Any = arrow.utcnow().timestamp
    ) -> Path:
        self.backup_path = Path(
            f"{self.config.app.name}-{int(timestamp)}-{class_name}.json"
        )
        return self.backup_path


class SqlBackupDates(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    today: arrow.Arrow = arrow.utcnow()
    white_list: list[int] = []
    black_list: list[int] = []
    dates: list[int] = []

    def __init__(self, dates: list[int] | None = None, **data: t.Any) -> None:
        super().__init__(**data)
        self.dates = sorted(dates, reverse=True) if dates else []
        self.process()

    def get_last_month_length(self) -> int:
        return monthrange(self.today.year, self.today.shift(months=-1).month)[1]

    def get_last_year_length(self) -> int:
        first_day = date(self.today.year, 1, 1)
        last_day = first_day - timedelta(days=1)
        return 366 if isleap(last_day.year) else 365

    def filter_dates(
        self, dates: list[t.Any], period: str = "week"
    ) -> t.Generator[t.Any, t.Any, t.Any]:
        reference = self.today.int_timestamp
        method_mapping: dict[str, t.Any] = {
            "week": lambda obj: getattr(obj, "isocalendar")()[1],
            "month": lambda obj: getattr(obj, "month"),
            "year": lambda obj: getattr(obj, "year"),
        }
        for dt in dates:
            comp_date = datetime.fromtimestamp(int(dt))
            method = method_mapping.get(period)
            if method is not None:
                comparison = method(comp_date)
                ref_date = datetime.fromtimestamp(reference)
                if comparison != method(ref_date):
                    reference = dt
                    yield dt

    def process(self) -> None:
        last_w = self.today.shift(days=-7).int_timestamp
        last_m = self.today.shift(days=-self.get_last_month_length()).int_timestamp
        last_y = self.today.shift(days=-self.get_last_year_length()).int_timestamp
        backups_week = []
        backups_month = []
        backups_year = []
        backups_older = []
        for timestamp in self.dates:
            if timestamp >= last_w:
                backups_week.append(timestamp)
            elif timestamp >= last_m:
                backups_month.append(timestamp)
            elif timestamp >= last_y:
                backups_year.append(timestamp)
            else:
                backups_older.append(timestamp)
        self.white_list.extend(
            chain(
                backups_week,
                self.filter_dates(backups_month),
                self.filter_dates(backups_year, "month"),
                self.filter_dates(backups_older, "year"),
            )
        )
        diff_as_list = [d for d in self.dates if d not in self.white_list]
        self.black_list.extend(sorted(diff_as_list, reverse=True))


class SqlBackup(SqlBackupDates, SqlBackupUtils):
    sql: Sql = depends()
    logger: Logger = depends()
    do_not_backup: list[str] = []
    models: list[t.Any] = []
    sure_delete: bool = False

    @staticmethod
    def show() -> list[t.Any]:
        schema = getattr(SQLModel.metadata, "schema", None) or []
        return [m for m in schema if isinstance(m, type) and issubclass(m, SQLModel)]

    def get_mapped_classes(self) -> list[t.Any]:
        self.add_subclasses(SQLModel)
        return self.models

    def add_subclasses(self, model: t.Any) -> None:
        if model.__subclasses__():
            for submodel in model.__subclasses__():
                self.add_subclasses(submodel)
        else:
            self.models.append(model)

    async def get_data(self) -> dict[str, t.Any]:
        data = {}
        async with self.sql.session() as session:
            for model in self.get_mapped_classes():
                query = select(model)
                results = session.exec(query)
                data[model.__name__] = results.model_dump_to_json()
        return data

    async def backup(self, class_name: str, data: t.Any, timestamp: int) -> None:
        path = self.get_path(class_name=class_name, timestamp=timestamp)
        await self.storage.sql.save(data, path)

    async def save(self) -> None:
        data = await self.get_data()
        timestamp = arrow.utcnow().int_timestamp
        for class_name in data.keys():
            path = self.get_path(class_name=class_name, timestamp=timestamp)
            self.logger.debug(f"Backing up - {path.name}")
            await self.backup(
                class_name=class_name, data=data[class_name], timestamp=timestamp
            )

    async def get_backups(self) -> list[int]:
        timestamps = self.get_timestamps()
        backups = {}
        for timestamp in timestamps:
            adate = arrow.Arrow.fromtimestamp(timestamp)
            date_formatted = adate.humanize()
            backups[timestamp] = []
            fps = [
                AsyncPath(b)
                async for b in self.storage.sql.list()
                if timestamp in b.name
            ]
            for fp in fps:
                if await self.storage.sql.exists(fp):
                    backups[timestamp].append(fp.name)
            adate_formatted = adate.format("MM-DD-YYYY HH:mm:ss ZZ")
            self.logger.debug(f"{date_formatted} ==> {adate_formatted}:")
        return sorted(backups, reverse=True)

    def get_last_backup(self) -> bool | int:
        timestamps = self.get_timestamps()
        if not timestamps:
            return False
        timestamp = max(timestamps)
        adate = arrow.Arrow.fromtimestamp(timestamp)
        self.logger.info(f"Last backup: {adate.format('MM-DD-YYYY HH:mm:ss ZZ')}")
        return timestamp

    async def restore_rows(self, path: AsyncPath) -> t.Any:
        contents = await self.storage.sql.read(path)
        pattern = "-(\\D+).json"
        model = search(pattern, path.name).group(1)
        if contents:
            loaded_model = create_model(
                model,
                __base__=SQLModel,
                __cls_kwargs__={"table": True},
                **t.cast(dict[str, t.Any], load.json(contents)),
            )
            if loaded_model:
                async with self.sql.session() as session:
                    for m in self.show():
                        m_name = m.__name__
                        if m_name == model:
                            await session.query(m).delete()
                    await session.commit()

    async def restore_backup(self, timestamp: int) -> None:
        paths = [
            AsyncPath(b.name)
            async for b in self.storage.sql.list()
            if timestamp in b.name
        ]
        when = arrow.Arrow.fromtimestamp(timestamp).humanize()
        self.logger.info(f"Restoring backup from {when}.....")
        for path in paths:
            self.logger.debug(f"\tRestoring:  {path.name}")
            await self.restore_rows(path)

    def delete_backups(self, delete_list: list[str]) -> None:
        delete_me = self.storage.sql.delete(delete_list)
        if delete_me:
            for name in delete_list:
                self.logger.debug(f"Deleted {name}")

    def clean(self) -> bool:
        cleaning = SqlBackupDates(self.get_timestamps())
        white_list = cleaning.white_list
        black_list = cleaning.black_list
        if not black_list:
            self.logger.debug("==> No backup to be deleted.")
            return True
        self.logger.debug(f"==> {len(white_list)} backups will be kept:")
        for timestamp in white_list:
            date_formatted = arrow.Arrow.fromtimestamp(timestamp).humanize()
            self.logger.debug(f"ID: {timestamp} (from {date_formatted})")
            if getattr(self.config.debug, "sql", False):
                for f in self.by_timestamp(str(timestamp)):
                    self.logger.debug(f)
        delete_list = []
        self.logger.debug(f"==> {len(black_list)} backups will be deleted:")
        for timestamp in black_list:
            date_formatted = arrow.Arrow.fromtimestamp(timestamp).humanize()
            self.logger.debug(f"ID: {timestamp} (from {date_formatted})")
            for f in self.by_timestamp(str(timestamp)):
                if getattr(self.config.debug, "database", False):
                    self.logger.debug(f)
                delete_list.append(f)
        self.delete_backups(delete_list)
        return True

    async def run(self) -> None:
        if not self.config.deployed:
            last_backup = self.get_last_backup()
            if getattr(self.config.debug, "database", False) and self.sure_delete:
                blobs = [b.name for b in self.storage.sql.list()]
                for b in blobs:
                    self.storage.sql.delete(b)
                last_backup = None
            if not last_backup:
                await self.save()
            elif (
                last_backup
                and getattr(self.config.debug, "database", False)
                and (not self.sure_delete)
            ):
                self.logger.info(f"Restoring last backup - {last_backup}")
                await self.restore_backup(last_backup)
            self.clean()
            self.logger.info("Backups complete.")


backup = SqlBackup()



"""Tests for the Gmail SMTP adapter."""

import typing as t
from collections.abc import AsyncGenerator, Callable
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from pydantic import SecretStr
from acb.adapters.smtp.gmail import Smtp as GmailSmtp
from acb.adapters.smtp.gmail import SmtpSettings as GmailSmtpSettings


@pytest.fixture
def mock_async_context_manager() -> Callable[
    ..., AbstractAsyncContextManager[MagicMock]
]:
    @asynccontextmanager
    async def _async_context_manager(
        *args: Any, **kwargs: Any
    ) -> AsyncGenerator[MagicMock]:
        yield MagicMock()

    return _async_context_manager


class TestGmailSmtpSettings:
    def test_init_basic(self) -> None:
        settings: GmailSmtpSettings = GmailSmtpSettings(
            name="gmail",
            domain="example.com",
            from_email="test@example.com",
            app_token="test_token",
            app_pwd="test_password",
            subdomain="mail",
        )
        assert settings.name == "gmail"  # type: ignore
        assert settings.domain == "example.com"
        assert settings.from_email == "test@example.com"  # type: ignore
        assert settings.app_token == "test_token"  # type: ignore
        assert settings.app_pwd == "test_password"  # type: ignore
        assert settings.subdomain == "mail"  # type: ignore

    def test_init_with_config(self) -> None:
        mock_config: MagicMock = MagicMock()
        mock_config.app.domain = "example.com"
        mock_config.app.title = "Example App"

        with patch("acb.depends.depends.__call__", return_value=mock_config):
            settings: GmailSmtpSettings = GmailSmtpSettings(
                domain="mail.example.com",
                subdomain="mail",
                default_from="info@example.com",
                default_from_name="Example App",
            )

            assert settings.domain == "mail.example.com"
            assert settings.default_from == "info@example.com"
            assert settings.default_from_name == "Example App"
            assert settings.token_uri == "https://oauth2.googleapis.com/token"
            assert "https://www.googleapis.com/auth/gmail.send" in settings.scopes
            assert settings.mx_servers is not None and len(settings.mx_servers) == 5
            assert "1 aspmx.l.google.com." in settings.mx_servers

        with patch("acb.depends.depends.__call__", return_value=mock_config):
            auth_settings: GmailSmtpSettings = GmailSmtpSettings(
                client_id="test-client-id",
                client_secret=SecretStr("test-client-secret"),
                refresh_token=SecretStr("test-refresh-token"),
                domain="mail.example.com",
                subdomain="mail",
                default_from="info@example.com",
                default_from_name="Example App",
            )
            assert auth_settings.client_id == "test-client-id"
            assert (
                auth_settings.client_secret.get_secret_value() == "test-client-secret"
            )
            assert (
                auth_settings.refresh_token.get_secret_value() == "test-refresh-token"
            )

    def test_validation(self) -> None:
        GmailSmtpSettings(
            name="gmail",
            domain="example.com",
            from_email="test@example.com",
            app_token="test_token",
            app_pwd="test_password",
        )


class TestGmailSmtp:
    @pytest.fixture
    def smtp(
        self,
        mock_async_context_manager: Callable[
            ..., AbstractAsyncContextManager[MagicMock]
        ],
    ) -> GmailSmtp:
        class TestableGmailSmtp(GmailSmtp):
            def __init__(self) -> None:
                super().__init__()
                self.connect = MagicMock()

        instance: GmailSmtp = TestableGmailSmtp()
        instance.config = MagicMock()
        instance.config.smtp = MagicMock()
        instance.config.smtp.domain = "example.com"
        instance.config.smtp.default_from = "info@example.com"
        instance.config.smtp.default_from_name = "Example App"
        instance.config.smtp.client_id = "test-client-id"
        instance.config.smtp.client_secret = MagicMock()
        instance.config.smtp.client_secret.get_secret_value.return_value = (
            "test-client-secret"
        )
        instance.config.smtp.scopes = ["https://www.googleapis.com/auth/gmail.send"]
        instance.config.smtp.mx_servers = [
            "1 aspmx.l.google.com.",
            "5 alt1.aspmx.l.google.com.",
        ]
        instance.config.smtp.forwards = {"admin": "admin@example.com"}
        instance.logger = MagicMock()
        instance.requests = MagicMock()

        if (
            not hasattr(instance, "_get_gmail_service")
            or instance._get_gmail_service is None
        ):
            instance._get_gmail_service = MagicMock(return_value=MagicMock())

        return instance

    async def test_get_gmail_service(self, smtp: GmailSmtp) -> None:
        def mock_get_gmail_service() -> t.Any:
            from acb.adapters.smtp.gmail import Credentials, build

            creds = Credentials(
                None,
                refresh_token="test-refresh-token",
                token_uri="https://oauth2.googleapis.com/token",
                client_id="test-client-id",
                client_secret="test-client-secret",
                scopes=["https://www.googleapis.com/auth/gmail.send"],
            )
            service = build("gmail", "v1", credentials=creds)
            return service

        smtp._get_gmail_service = mock_get_gmail_service

        with (
            patch("acb.adapters.smtp.gmail.Credentials") as mock_credentials,
            patch("acb.adapters.smtp.gmail.build") as mock_build,
        ):
            mock_service = MagicMock()
            mock_build.return_value = mock_service

            service = smtp._get_gmail_service()

            mock_credentials.assert_called_once_with(
                None,
                refresh_token="test-refresh-token",
                token_uri="https://oauth2.googleapis.com/token",
                client_id="test-client-id",
                client_secret="test-client-secret",
                scopes=["https://www.googleapis.com/auth/gmail.send"],
            )
            mock_build.assert_called_once_with(
                "gmail", "v1", credentials=mock_credentials.return_value
            )
            assert service == mock_service

    @pytest.mark.asyncio
    async def test_get_response(self, smtp: GmailSmtp) -> None:
        # Create mock logger
        mock_logger = MagicMock()
        smtp.logger = mock_logger

        result = await smtp.get_response("get", domain="example.com")

        mock_logger.debug.assert_called_once()
        assert result["message"] == "success"
        assert result["status"] == "ok"

    @pytest.mark.asyncio
    async def test_list_domains(self, smtp: GmailSmtp) -> None:
        domains = await smtp.list_domains()
        assert domains == ["example.com"]

    @pytest.mark.asyncio
    async def test_get_domain(self, smtp: GmailSmtp) -> None:
        domain_info = await smtp.get_domain("example.com")
        assert domain_info is not None and domain_info["domain"] == "example.com"
        assert (
            domain_info is not None and len(domain_info["receiving_dns_records"]) == 2
        )
        assert (
            domain_info is not None
            and domain_info["receiving_dns_records"][0]["priority"] == "1"
        )
        assert (
            domain_info is not None
            and domain_info["receiving_dns_records"][0]["value"]
            == "aspmx.l.google.com."
        )
        assert domain_info is not None and len(domain_info["sending_dns_records"]) == 2
        assert (
            domain_info is not None
            and "v=spf1" in domain_info["sending_dns_records"][0]["value"]
        )
        assert (
            domain_info is not None
            and "v=DMARC1" in domain_info["sending_dns_records"][1]["value"]
        )

    @pytest.mark.asyncio
    async def test_create_domain(self, smtp: GmailSmtp) -> None:
        # Create mock logger
        mock_logger = MagicMock()
        smtp.logger = mock_logger

        result = await smtp.create_domain("example.com")

        mock_logger.info.assert_called_once()
        assert result["message"] == "Domain configured for Gmail"
        assert result["domain"] == "example.com"

    @pytest.mark.asyncio
    async def test_delete_domain(self, smtp: GmailSmtp) -> None:
        # Create mock logger
        mock_logger = MagicMock()
        smtp.logger = mock_logger

        result = await smtp.delete_domain("example.com")

        mock_logger.info.assert_called_once()
        assert result["message"] == "Domain deletion simulated"
        assert result["domain"] == "example.com"

    @pytest.mark.asyncio
    async def test_get_dns_records(self, smtp: GmailSmtp) -> None:
        with patch("acb.adapters.smtp.gmail.debug") as mock_debug:
            records = await smtp.get_dns_records("example.com")

            assert len(records) == 3
            assert records[0].type == "MX"
            assert records[1].type == "TXT"
            assert records[2].type == "TXT"
            mock_debug.assert_called_once()

    @pytest.mark.asyncio
    async def test_create_dns_records(self, smtp: GmailSmtp) -> None:
        mock_dns = AsyncMock()

        with patch.object(smtp, "get_dns_records", AsyncMock()) as mock_get_records:
            mock_records = [MagicMock(), MagicMock()]
            mock_get_records.return_value = mock_records

            # Create mock logger
            mock_logger = MagicMock()
            smtp.logger = mock_logger

            await smtp.create_dns_records(dns=mock_dns)

            mock_get_records.assert_called_once_with("example.com")
            mock_dns.create_records.assert_called_once_with(mock_records)
            mock_logger.info.assert_called_once()

    @pytest.mark.asyncio
    async def test_list_routes(self, smtp: GmailSmtp) -> None:
        with patch.object(smtp, "_get_gmail_service") as mock_get_service:
            mock_service = MagicMock()
            mock_get_service.return_value = mock_service

            mock_list = MagicMock()
            mock_service.users().settings().forwardingAddresses().list.return_value = (
                mock_list
            )
            mock_list.execute.return_value = {
                "forwardingAddresses": [{"forwardingEmail": "admin@example.com"}]
            }

            routes = await smtp.list_routes()

            assert len(routes) == 1
            assert routes[0]["id"] == "admin@example.com"
            assert "match_recipient" in routes[0]["expression"]
            assert "forward" in routes[0]["actions"][0]

    @pytest.mark.asyncio
    async def test_delete_route(self, smtp: GmailSmtp) -> None:
        # Create mock logger BEFORE calling the method
        mock_logger = MagicMock()
        smtp.logger = mock_logger

        with patch.object(smtp, "_get_gmail_service") as mock_get_service:
            mock_service = MagicMock()
            mock_get_service.return_value = mock_service

            mock_delete = MagicMock()
            mock_service.users().settings().forwardingAddresses().delete.return_value = mock_delete

            route = {"id": "admin@example.com", "description": "Admin route"}
            response = await smtp.delete_route(route)

            mock_service.users().settings().forwardingAddresses().delete.assert_called_once_with(
                userId="me", forwardingEmail="admin@example.com"
            )

            assert response.status_code == 200
            mock_logger.info.assert_called_once()

    @pytest.mark.asyncio
    async def test_create_route(self, smtp: GmailSmtp) -> None:
        # Create mock logger BEFORE calling the method
        mock_logger = MagicMock()
        smtp.logger = mock_logger

        with patch.object(smtp, "_get_gmail_service") as mock_get_service:
            mock_service = MagicMock()
            mock_get_service.return_value = mock_service

            mock_create = MagicMock()
            mock_service.users().settings().forwardingAddresses().create.return_value = mock_create

            mock_update = MagicMock()
            mock_service.users().settings().updateAutoForwarding.return_value = (
                mock_update
            )

            result = await smtp.create_route("admin", "admin@example.com")

            mock_service.users().settings().forwardingAddresses().create.assert_called_once_with(
                userId="me", body={"forwardingEmail": "admin@example.com"}
            )
            mock_service.users().settings().updateAutoForwarding.assert_called_once()

            assert result["message"] == "Forwarding created"
            mock_logger.info.assert_called()

    @pytest.mark.asyncio
    async def test_send_email(self, smtp: GmailSmtp) -> None:
        # Create mock logger BEFORE calling the method
        mock_logger = MagicMock()
        smtp.logger = mock_logger

        with (
            patch.object(smtp, "_get_gmail_service") as mock_get_service,
            patch("acb.adapters.smtp.gmail.MIMEMultipart") as mock_mime_multipart,
            patch("acb.adapters.smtp.gmail.MIMEText") as mock_mime_text,
            patch("acb.adapters.smtp.gmail.base64") as mock_base64,
        ):
            mock_service = MagicMock()
            mock_get_service.return_value = mock_service

            mock_message = MagicMock()
            mock_mime_multipart.return_value = mock_message

            mock_send = MagicMock()
            mock_service.users().messages().send.return_value = mock_send
            mock_send.execute.return_value = {"id": "message-id-123"}

            mock_base64.urlsafe_b64encode.return_value.decode.return_value = (
                "encoded-message"
            )

            result = await smtp.send_email(
                to="user@example.com",
                subject="Test Subject",
                body="Test Body",
            )

            mock_mime_multipart.assert_called_once_with("alternative")
            mock_message.__setitem__.assert_any_call("to", "user@example.com")
            mock_message.__setitem__.assert_any_call("subject", "Test Subject")
            mock_mime_text.assert_called_once_with("Test Body", "plain")
            mock_message.attach.assert_called_once_with(mock_mime_text.return_value)

            mock_service.users().messages().send.assert_called_once_with(
                userId="me", body={"raw": "encoded-message"}
            )

            assert result["id"] == "message-id-123"
            assert result["status"] == "sent"
            mock_logger.info.assert_called_once()

    @pytest.mark.asyncio
    async def test_init(self, smtp: GmailSmtp) -> None:
        with (
            patch.object(smtp, "create_dns_records", AsyncMock()) as mock_create_dns,
            patch.object(smtp, "delete_routes", AsyncMock()) as mock_delete_routes,
            patch.object(smtp, "create_routes", AsyncMock()) as mock_create_routes,
        ):
            await smtp.init()

            mock_create_dns.assert_called_once()
            mock_delete_routes.assert_called_once()
            mock_create_routes.assert_called_once()



import typing as t

import niquests
from pydantic import SecretStr
from acb.depends import depends

from ._base import RequestsBase, RequestsBaseSettings


class RequestsSettings(RequestsBaseSettings):
    base_url: str = ""
    timeout: int = 10
    auth: tuple[str, SecretStr] | None = None


class Requests(RequestsBase):
    def __init__(self, **kwargs: t.Any) -> None:
        super().__init__()

    async def _create_client(self) -> "niquests.AsyncSession":
        session = niquests.AsyncSession()
        if self.config.requests.base_url:
            session.base_url = self.config.requests.base_url
        return session

    async def get_client(self) -> "niquests.AsyncSession":
        return await self._ensure_client()

    @property
    def client(self) -> "niquests.AsyncSession":
        if self._client is None:
            raise RuntimeError("Client not initialized. Call get_client() first.")
        return self._client

    async def get(
        self,
        url: str,
        timeout: int = 10,
        params: dict[str, t.Any] | None = None,
        headers: dict[str, str] | None = None,
        cookies: dict[str, str] | None = None,
    ) -> t.Any:
        client = await self.get_client()
        response = await client.get(
            url, timeout=timeout, params=params, headers=headers, cookies=cookies
        )
        response.raise_for_status()
        return response

    async def post(
        self,
        url: str,
        data: dict[str, t.Any] | None = None,
        timeout: int = 10,
        json: dict[str, t.Any] | None = None,
    ) -> t.Any:
        client = await self.get_client()
        response = await client.post(url, data=data, json=json, timeout=timeout)
        response.raise_for_status()
        return response

    async def put(
        self,
        url: str,
        data: dict[str, t.Any] | None = None,
        timeout: int = 10,
        json: dict[str, t.Any] | None = None,
    ) -> t.Any:
        client = await self.get_client()
        response = await client.put(url, data=data, json=json, timeout=timeout)
        response.raise_for_status()
        return response

    async def delete(self, url: str, timeout: int = 10) -> t.Any:
        client = await self.get_client()
        response = await client.delete(url, timeout=timeout)
        response.raise_for_status()
        return response

    async def patch(
        self,
        url: str,
        timeout: int = 10,
        data: dict[str, t.Any] | None = None,
        json: dict[str, t.Any] | None = None,
    ) -> t.Any:
        client = await self.get_client()
        response = await client.patch(url, timeout=timeout, data=data, json=json)
        response.raise_for_status()
        return response

    async def head(self, url: str, timeout: int = 10) -> t.Any:
        client = await self.get_client()
        response = await client.head(url, timeout=timeout)
        response.raise_for_status()
        return response

    async def options(self, url: str, timeout: int = 10) -> t.Any:
        client = await self.get_client()
        response = await client.options(url, timeout=timeout)
        response.raise_for_status()
        return response

    async def request(
        self,
        method: str,
        url: str,
        data: dict[str, t.Any] | None = None,
        json: dict[str, t.Any] | None = None,
        timeout: int = 10,
    ) -> t.Any:
        client = await self.get_client()
        response = await client.request(
            method, url, data=data, json=json, timeout=timeout
        )
        response.raise_for_status()
        return response

    async def close(self) -> None:
        if self._client is not None:
            await self._client.close()

    async def init(self) -> None:
        self.logger.debug("Niquests adapter initialized")


depends.set(Requests)



"""Tests for the MongoDB NoSQL adapter."""

from collections.abc import AsyncGenerator, Callable
from contextlib import asynccontextmanager
from types import TracebackType
from typing import Any
from unittest.mock import MagicMock

import pytest
from pydantic import SecretStr
from acb.adapters.nosql.mongodb import Nosql as MongoDBNosql
from acb.adapters.nosql.mongodb import NosqlSettings as MongoDBSettings


@pytest.fixture
async def mock_async_context_manager() -> Callable[[Any | None], Any]:
    class MockAsyncContextManager:
        def __init__(self, mock_obj: Any | None = None) -> None:
            self.mock_obj = mock_obj or MagicMock()

        async def __aenter__(self) -> Any:
            return self.mock_obj

        async def __aexit__(
            self,
            exc_type: type[BaseException] | None,
            exc_val: BaseException | None,
            exc_tb: TracebackType | None,
        ) -> None:
            pass

    def _mock_async_context_manager(
        mock_obj: Any | None = None,
    ) -> MockAsyncContextManager:
        return MockAsyncContextManager(mock_obj)

    return _mock_async_context_manager


class TestMongoDBSettings:
    def test_init(self) -> None:
        settings = MongoDBSettings(
            host=SecretStr("localhost"),
            port=27017,
            database="test_db",
            user=SecretStr("test_user"),
            password=SecretStr("test_password"),
        )

        assert settings.host.get_secret_value() == "localhost"
        assert settings.port == 27017
        assert settings.database == "test_db"
        assert settings.user.get_secret_value() == "test_user"
        assert settings.password.get_secret_value() == "test_password"

        conn_str = settings.connection_string
        assert conn_str is not None
        assert "mongodb://test_user:test_password@localhost:27017/test_db" == conn_str

        settings = MongoDBSettings(
            host=SecretStr("localhost"),
            database="test_db",
        )

        assert settings.host.get_secret_value() == "localhost"
        assert settings.port == 27017
        assert settings.database == "test_db"
        assert settings.user is None
        assert settings.password is None

        conn_str = settings.connection_string
        assert conn_str is not None
        assert "mongodb://localhost:27017/test_db" == conn_str


class TestMongoDB:
    @pytest.fixture
    def nosql(self) -> MongoDBNosql:
        nosql = MongoDBNosql()
        nosql.config = MagicMock()
        nosql.logger = MagicMock()

        nosql.config.nosql.collection_prefix = ""

        return nosql

    @pytest.mark.asyncio
    async def test_init(self, nosql: MongoDBNosql) -> None:
        mongodb_settings = MongoDBSettings(
            host=SecretStr("localhost"),
            port=27017,
            database="test_db",
        )

        nosql.config.nosql.get.return_value = mongodb_settings

        original_init = nosql.init
        mock_logger = MagicMock()
        nosql.logger = mock_logger

        async def mock_init() -> None:
            nosql.logger.info("Initializing MongoDB connection")
            nosql.logger.info("MongoDB connection initialized successfully")
            return None

        nosql.init = mock_init

        await nosql.init()
        assert mock_logger.info.call_count == 2

        nosql.init = original_init

    @pytest.mark.asyncio
    async def test_init_error(self, nosql: MongoDBNosql) -> None:
        mongodb_settings = MongoDBSettings(
            host=SecretStr("localhost"),
            port=27017,
            database="test_db",
        )

        nosql.config.nosql.get.return_value = mongodb_settings

        original_init = nosql.init
        mock_logger = MagicMock()
        nosql.logger = mock_logger

        async def mock_init_error() -> None:
            nosql.logger.info("Initializing MongoDB connection")
            nosql.logger.error(
                "Failed to initialize MongoDB connection: Connection error"
            )
            raise Exception("Connection error")

        nosql.init = mock_init_error

        with pytest.raises(Exception) as excinfo:
            await nosql.init()
        assert "Connection error" in str(excinfo.value)
        assert mock_logger.error.call_count == 1

        nosql.init = original_init

    @pytest.mark.asyncio
    async def test_find(self, nosql: MongoDBNosql) -> None:
        original_find = nosql.find

        async def mock_find(
            collection: str, filter: dict[str, Any], **kwargs: Any
        ) -> list[dict[str, Any]]:
            assert collection == "test_collection"
            assert filter == {"test": "test"}
            return [{"_id": "test_id", "data": "test_data"}]

        nosql.find = mock_find

        result = await nosql.find("test_collection", {"test": "test"})
        assert result == [{"_id": "test_id", "data": "test_data"}]

        nosql.find = original_find

    @pytest.mark.asyncio
    async def test_find_one(self, nosql: MongoDBNosql) -> None:
        original_find_one = nosql.find_one

        async def mock_find_one(
            collection: str, filter: dict[str, Any], **kwargs: Any
        ) -> dict[str, Any]:
            assert collection == "test_collection"
            assert filter == {"_id": "test_id"}
            return {"_id": "test_id", "data": "test_data"}

        nosql.find_one = mock_find_one

        result = await nosql.find_one("test_collection", {"_id": "test_id"})

        assert result == {"_id": "test_id", "data": "test_data"}

        nosql.find_one = original_find_one

    @pytest.mark.asyncio
    async def test_find_many(self, nosql: MongoDBNosql) -> None:
        original_find = nosql.find

        async def mock_find(
            collection: str, filter: dict[str, Any], **kwargs: Any
        ) -> list[dict[str, Any]]:
            if collection == "test_collection" and filter == {"status": "active"}:
                return [
                    {"_id": "1", "data": "test_data_1", "status": "active"},
                    {"_id": "2", "data": "test_data_2", "status": "active"},
                ]
            return await original_find(collection, filter, **kwargs)

        nosql.find = mock_find

        results = await nosql.find("test_collection", {"status": "active"})

        assert len(results) == 2
        assert results[0]["_id"] == "1"
        assert results[1]["_id"] == "2"

        nosql.find = original_find

    @pytest.mark.asyncio
    async def test_insert_one(self, nosql: MongoDBNosql) -> None:
        original_insert_one = nosql.insert_one

        async def mock_insert_one(
            collection: str, document: dict[str, Any], **kwargs: Any
        ) -> dict[str, Any]:
            assert collection == "test_collection"
            assert document == {"data": "test_data"}

            return {"inserted_id": "test_id"}

        nosql.insert_one = mock_insert_one

        result = await nosql.insert_one("test_collection", {"data": "test_data"})

        assert result["inserted_id"] == "test_id"

        nosql.insert_one = original_insert_one

    @pytest.mark.asyncio
    async def test_transaction(self, nosql: MongoDBNosql) -> None:
        original_transaction = nosql.transaction

        @asynccontextmanager
        async def mock_transaction() -> AsyncGenerator[None]:
            nosql.logger.info("Starting transaction")
            yield
            nosql.logger.info("Ending transaction")

        nosql.transaction = mock_transaction

        async with nosql.transaction():
            nosql.logger.info("Inside transaction")

        nosql.transaction = original_transaction

    def test_init_with_connection_string(self) -> None:
        connection_string = "mongodb://test_user:test_password@localhost:27017/test_db"
        nosql = MongoDBNosql()

        nosql.config = MagicMock()
        nosql.config.nosql.connection_string = connection_string
        nosql.connection_string = connection_string

        assert nosql.connection_string == connection_string

    def test_init_with_host_and_db(self) -> None:
        expected_connection = "mongodb://localhost:27017/test_db"
        nosql = MongoDBNosql()

        nosql.config = MagicMock()
        nosql.config.nosql.connection_string = expected_connection
        nosql.connection_string = expected_connection

        assert nosql.connection_string == expected_connection



import logging
import os
import sys
import typing as t
from inspect import currentframe

import nest_asyncio
from aioconsole import aprint
from loguru._logger import Core as _Core
from loguru._logger import Logger as _Logger

from .config import Config, Settings, debug
from .depends import depends

nest_asyncio.apply()


class LoggerSettings(Settings):
    verbose: bool = False
    deployed_level: str = "WARNING"
    log_level: str | None = "INFO"
    serialize: bool | None = False
    format: dict[str, str] | None = dict(
        time="<b><e>[</e> <w>{time:YYYY-MM-DD HH:mm:ss.SSS}</w> <e>]</e></b>",
        level=" <level>{level:>8}</level>",
        sep=" <b><w>in</w></b> ",
        name="<b>{extra[mod_name]:>20}</b>",
        line="<b><e>[</e><w>{line:^5}</w><e>]</e></b>",
        message="  <level>{message}</level>",
    )
    level_per_module: dict[str, str | None] | None = {}
    level_colors: dict[str, str] | None = {}
    settings: dict[str, t.Any] | None = {}

    def __init__(self, **values: t.Any) -> None:
        super().__init__(**values)
        self.settings = dict(
            format="".join(self.format.values() if self.format else []),
            enqueue=True,
            backtrace=False,
            catch=False,
            serialize=self.serialize,
            diagnose=False,
            colorize=True,
        )


depends.get(Config).logger = LoggerSettings()


@t.runtime_checkable
class LoggerProtocol(t.Protocol):
    def debug(self, msg: str, *args: t.Any, **kwargs: t.Any) -> None: ...
    def info(self, msg: str, *args: t.Any, **kwargs: t.Any) -> None: ...
    def warning(self, msg: str, *args: t.Any, **kwargs: t.Any) -> None: ...
    def error(self, msg: str, *args: t.Any, **kwargs: t.Any) -> None: ...
    def init(self) -> None: ...


class LoggerBase(LoggerProtocol):
    config: Config = depends()


class Logger(_Logger, LoggerBase):
    @depends.inject
    def __init__(self) -> None:
        super().__init__(
            core=_Core(),
            exception=None,
            depth=0,
            record=False,
            lazy=False,
            colors=False,
            raw=False,
            capture=True,
            patchers=[],
            extra={},
        )

    @staticmethod
    async def async_sink(message: str) -> None:
        await aprint(message, end="")

    def init(self) -> None:
        if self._is_testing_mode():
            self._configure_for_testing()
            return
        self._configure_logger()
        self._setup_level_colors()
        self._log_debug_levels()
        self._log_app_info()

    def _is_testing_mode(self) -> bool:
        return (
            "pytest" in sys.modules or os.getenv("TESTING", "False").lower() == "true"
        )

    def _configure_for_testing(self) -> None:
        self.remove()
        self.configure(handlers=[])

    def _configure_logger(self) -> None:
        self.remove()
        self.configure(
            patcher=lambda record: record["extra"].update(
                mod_name=self._patch_name(record)
            )
        )
        self.config.logger.log_level = (
            self.config.logger.deployed_level.upper()
            if self.config.deployed or self.config.debug.production
            else self.config.logger.log_level
        )
        self.config.logger.level_per_module = {
            m: "DEBUG" if v else self.config.logger.log_level for m, v in debug.items()
        }
        self._add_logger_sink()

    def _patch_name(self, record: dict[str, t.Any]) -> str:
        mod_parts = record["name"].split(".")
        mod_name = ".".join(mod_parts[:-1])
        if len(mod_parts) > 3:
            mod_name = ".".join(mod_parts[1:-1])
        return mod_name.replace("_sdk", "")

    def _filter_by_module(self, record: dict[str, t.Any]) -> bool:
        try:
            name = record["name"].split(".")[-2]
        except IndexError:
            name = record["name"]
        level_ = self.config.logger.log_level
        if name in self.config.logger.level_per_module:
            level_ = self.config.logger.level_per_module[name]
        try:
            levelno_ = self.level(level_).no
        except ValueError:
            raise ValueError(
                f"The filter dict contains a module '{name}' associated to a level  name which does not exist: '{level_}'"
            )
        if level_ is False:
            return False
        return record["level"].no >= levelno_

    def _add_logger_sink(self) -> None:
        try:
            self.add(
                self.async_sink,
                filter=t.cast(t.Any, self._filter_by_module),
                **self.config.logger.settings,
            )
        except ValueError as e:
            if "event loop is required" in str(e):
                self._add_sync_sink()
            else:
                raise

    def _add_sync_sink(self) -> None:
        self.add(
            lambda msg: print(msg, end=""),
            filter=t.cast(t.Any, self._filter_by_module),
            **{k: v for k, v in self.config.logger.settings.items() if k != "enqueue"},
        )

    def _setup_level_colors(self) -> None:
        for level, color in self.config.logger.level_colors.items():
            self.level(level.upper(), color=f"[{color}]")

    def _log_debug_levels(self) -> None:
        if self.config.debug.logger:
            self.debug("debug")
            self.info("info")
            self.warning("warning")
            self.error("error")
            self.critical("critical")

    def _log_app_info(self) -> None:
        self.info(f"App path: {self.config.root_path}")
        self.info(f"App deployed: {self.config.deployed}")


depends.set(Logger)
depends.get(Logger).init()


class InterceptHandler(logging.Handler):
    @depends.inject
    def emit(self, record: logging.LogRecord, logger: Logger = depends()) -> None:
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno
        frame, depth = (currentframe(), 0)
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1
        logger.opt(depth=depth, exception=record.exc_info).log(
            level, record.getMessage()
        )


logging.basicConfig(handlers=[InterceptHandler()], level=0, force=True)



import asyncio
import sys
import typing as t
from contextvars import ContextVar
from enum import Enum
from functools import cached_property
from pathlib import Path
from secrets import token_bytes, token_urlsafe
from string import punctuation
from weakref import WeakKeyDictionary

import nest_asyncio
import rich.repr
from anyio import Path as AsyncPath
from inflection import titleize, underscore
from pydantic import BaseModel, ConfigDict, Field, SecretStr, field_validator
from pydantic._internal._utils import deep_update
from pydantic.dataclasses import dataclass
from pydantic.fields import FieldInfo
from pydantic_settings import SettingsConfigDict, SettingsError

from .actions.encode import dump, load
from .adapters import (
    _deployed,
    _testing,
    adapter_registry,
    get_adapter,
    import_adapter,
    root_path,
    secrets_path,
    settings_path,
    tmp_path,
)
from .depends import depends

if not _testing:
    nest_asyncio.apply()
project: str = ""
app_name: str = ""
debug: dict[str, bool] = {}
_app_secrets: ContextVar[set[str]] = ContextVar("_app_secrets", default=set())

_config_initialized: bool = False
_library_usage_mode: bool = False


def _is_build_tool_execution() -> bool:
    return any(cmd in sys.argv[0] for cmd in ("pip", "setup.py", "build", "install"))


def _should_skip_frame(filename: str) -> bool:
    if any(skip in filename for skip in ("/site-packages/", "importlib", "_bootstrap")):
        return True
    return "acb" in filename


def _is_project_root(directory: Path) -> bool:
    return (directory / "pyproject.toml").exists() or (directory / "setup.py").exists()


def _get_frame_stack() -> list[t.Any]:
    import inspect

    frame = inspect.currentframe()
    if frame is None:
        return []
    frames = []
    caller_frame = frame
    while caller_frame:
        caller_frame = caller_frame.f_back
        if caller_frame is None:
            break
        frames.append(caller_frame)
    return frames


def _should_skip_caller_directory(caller_dir: Path) -> bool:
    return caller_dir.name in ("acb", "test", "tests")


def _find_project_root_in_path(caller_dir: Path) -> bool:
    for potential_root in [caller_dir] + list(caller_dir.parents):
        if _is_project_root(potential_root):
            return True
    return False


def _process_frame_for_project(caller_frame: t.Any) -> bool:
    filename = caller_frame.f_code.co_filename
    if _should_skip_frame(filename):
        return False
    caller_dir = Path(filename).parent
    if _should_skip_caller_directory(caller_dir):
        return False

    return _find_project_root_in_path(caller_dir)


def _check_caller_frame_for_project() -> bool:
    from contextlib import suppress

    with suppress(Exception):
        frames = _get_frame_stack()
        for caller_frame in frames:
            if _process_frame_for_project(caller_frame):
                return True
    return False


def _detect_library_usage() -> bool:
    if _is_build_tool_execution():
        return True
    if _testing or "pytest" in sys.modules:
        return False
    return _check_caller_frame_for_project()


def _is_pytest_test_context() -> bool:
    if "pytest" not in sys.modules:
        return False
    from contextlib import suppress

    with suppress(Exception):
        import inspect

        for frame_info in inspect.stack():
            filename = frame_info.filename
            if "acb/tests/" in filename or filename.endswith(
                "/acb/tests/test_config.py"
            ):
                return True
    return False


def _is_main_module_local() -> bool:
    main_module = sys.modules.get("__main__")
    if (
        not main_module
        or not hasattr(main_module, "__file__")
        or main_module.__file__ is None
    ):
        return False
    main_file = Path(main_module.__file__)
    cwd = Path.cwd()
    return main_file.parent == cwd or "acb" in str(main_file)


def _should_initialize_eagerly() -> bool:
    if _is_pytest_test_context():
        return True
    if _testing:
        return True
    if _detect_library_usage():
        return False
    return _is_main_module_local()


_library_usage_mode = _detect_library_usage()


class Platform(str, Enum):
    aws = "aws"
    gcp = "gcp"
    azure = "azure"
    cloudflare = "cloudflare"


async def get_version() -> str:
    pyproject_toml = root_path.parent / "pyproject.toml"
    if await pyproject_toml.exists():
        data = await load.toml(pyproject_toml)
        version = data.get("project", {}).get("version", "0.1.0")
        return version
    return "0.1.0"


def get_version_default() -> str:
    return asyncio.run(get_version())


def gen_password(size: int = 10) -> str:
    return token_urlsafe(size)


class PydanticSettingsProtocol(t.Protocol):
    adapter_name: str
    secrets_path: AsyncPath
    settings_cls: type["Settings"]
    model_config: SettingsConfigDict

    def __init__(
        self, settings_cls: type["Settings"], secrets_path: AsyncPath = ...
    ) -> None: ...
    def get_model_secrets(self) -> dict[str, FieldInfo]: ...
    async def __call__(self) -> dict[str, t.Any]: ...
    def __repr__(self) -> str: ...


class PydanticSettingsSource:
    adapter_name: str = "app"
    settings_cls: type["Settings"]
    model_config = SettingsConfigDict(arbitrary_types_allowed=True)

    def __init__(
        self, settings_cls: type["Settings"], secrets_path: AsyncPath | None = None
    ) -> None:
        from acb.adapters import secrets_path as sp

        self.settings_cls = settings_cls
        self.adapter_name = underscore(
            self.settings_cls.__name__.replace("Settings", "")
        )
        self.secrets_path = secrets_path if secrets_path is not None else sp
        self.config = settings_cls.model_config

    def get_model_secrets(self) -> dict[str, FieldInfo]:
        return {
            "_".join((self.adapter_name, n)): v
            for n, v in self.settings_cls.model_fields.items()
            if v.annotation is SecretStr
        }

    async def __call__(self) -> dict[str, t.Any]:
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(secrets_path={self.secrets_path!r})"


class InitSettingsSource(PydanticSettingsSource):
    def __init__(
        self, settings_cls: type["Settings"], init_kwargs: dict[str, t.Any]
    ) -> None:
        self.init_kwargs = init_kwargs
        super().__init__(settings_cls=settings_cls)

    async def __call__(self) -> dict[str, t.Any]:
        return self.init_kwargs


class FileSecretSource(PydanticSettingsSource):
    async def get_field_value(self, field_name: str) -> SecretStr | None:
        path = self.secrets_path / field_name
        if await path.is_file():
            return SecretStr((await path.read_text()).strip())
        return None

    def _get_test_secret_data(self) -> dict[str, t.Any]:
        data: dict[str, t.Any] = {}
        model_secrets = self.get_model_secrets()
        for field_key, field_info in model_secrets.items():
            field_name = field_key.removeprefix(f"{self.adapter_name}_")
            if hasattr(field_info, "default") and field_info.default is not None:
                data[field_name] = field_info.default
            else:
                data[field_name] = SecretStr(f"test_secret_for_{field_name}")
        return data

    async def _ensure_secrets_directory(self) -> None:
        self.secrets_path = await AsyncPath(self.secrets_path).expanduser()
        if not await self.secrets_path.exists():
            await self.secrets_path.mkdir(parents=True, exist_ok=True)

    async def _process_secret_field(
        self, field_key: str, data: dict[str, t.Any]
    ) -> None:
        cleaned_field_key = field_key.removeprefix(f"{app_name}_")
        if cleaned_field_key in _app_secrets.get():
            return

        try:
            field_value = await self.get_field_value(cleaned_field_key)
        except Exception as e:
            raise SettingsError(
                f"Error getting value for field '{cleaned_field_key}' from source '{self.__class__.__name__}: {e}'"
            )

        if field_value is not None:
            field_name = cleaned_field_key.removeprefix(f"{self.adapter_name}_")
            data[field_name] = field_value
            _app_secrets.get().add(cleaned_field_key)

    async def __call__(self) -> dict[str, t.Any]:
        if _testing or _library_usage_mode:
            return self._get_test_secret_data()
        await self._ensure_secrets_directory()
        data: dict[str, t.Any] = {}
        model_secrets = self.get_model_secrets()
        for field_key in model_secrets:
            await self._process_secret_field(field_key, data)
        return data


class ManagerSecretSource(PydanticSettingsSource):
    @cached_property
    def secret_manager(self):
        secret = depends.get()
        return secret

    async def load_secrets(self) -> t.Any:
        if _testing or _library_usage_mode:
            return await self._load_test_secrets()
        return await self._load_production_secrets()

    async def _load_test_secrets(self) -> dict[str, t.Any]:
        data = {}
        adapter_secrets = self.get_model_secrets()
        for field_key, field_info in adapter_secrets.items():
            field_name = field_key.removeprefix(f"{self.adapter_name}_")
            if hasattr(field_info, "default") and field_info.default is not None:
                data[field_name] = field_info.default
            else:
                data[field_name] = SecretStr(f"test_secret_for_{field_name}")
        return data

    async def _load_production_secrets(self) -> dict[str, t.Any]:
        data: dict[str, t.Any] = {}
        adapter_secrets = self.get_model_secrets()
        missing_secrets = {
            n: v for n, v in adapter_secrets.items() if n not in _app_secrets.get()
        }
        if missing_secrets and self.secret_manager:
            await self._process_missing_secrets(missing_secrets, data)
        return data

    async def _process_missing_secrets(
        self, missing_secrets: dict[str, t.Any], data: dict[str, t.Any]
    ) -> None:
        manager_secrets = await self.secret_manager.list(self.adapter_name)
        for field_key, field_value in missing_secrets.items():
            await self._process_secret_field(
                field_key, field_value, manager_secrets, data
            )

    async def _process_secret_field(
        self,
        field_key: str,
        field_value: t.Any,
        manager_secrets: list[str],
        data: dict[str, t.Any],
    ) -> None:
        stored_field_key = "_".join((app_name, field_key))
        secret_path = self.secrets_path / field_key

        if not await secret_path.exists() and field_key not in manager_secrets:
            await self.secret_manager.create(
                stored_field_key, field_value.default.get_secret_value()
            )

        secret = await self.secret_manager.get(stored_field_key)
        await secret_path.write_text(secret)
        field_name = field_key.removeprefix(f"{self.adapter_name}_")
        data[field_name] = SecretStr(secret)
        _app_secrets.get().add(field_key)

    async def __call__(self) -> t.Any:
        data = await self.load_secrets()
        return data


class YamlSettingsSource(PydanticSettingsSource):
    def _get_default_settings(self) -> dict[str, t.Any]:
        return {
            name: info.default
            for name, info in self.settings_cls.model_fields.items()
            if info.annotation is not SecretStr
        }

    def _get_dump_settings(self) -> dict[str, t.Any]:
        return {
            name: info.default
            for name, info in self.settings_cls.model_fields.items()
            if info.annotation is not SecretStr
            and "Optional" not in str(info.annotation)
        }

    def _update_global_variables(self, settings: dict[str, t.Any]) -> None:
        global project, app_name, debug
        if self.adapter_name == "debug":
            debug = settings
        elif self.adapter_name == "app":
            if _testing or _library_usage_mode:
                project = "test_project" if _testing else "library_project"
                app_name = "test_app" if _testing else "library_app"
            else:
                project = settings.get("project", "")
                app_name = settings.get("name", "")

    def _process_debug_settings(
        self, yml_settings: dict[str, t.Any]
    ) -> dict[str, t.Any]:
        for adapter in adapter_registry.get():
            if adapter.category not in (yml_settings.keys() or ("config", "logger")):
                yml_settings[adapter.category] = False
        return yml_settings

    async def _handle_testing_mode(self) -> dict[str, t.Any]:
        default_settings = self._get_default_settings()
        self._update_global_variables(default_settings)
        return default_settings

    async def _create_yml_file_if_needed(self, yml_path: AsyncPath) -> None:
        if not await yml_path.exists() and not _deployed:
            dump_settings = self._get_dump_settings()
            await dump.yaml(dump_settings, yml_path)

    async def load_yml_settings(self) -> t.Any:
        if self.adapter_name == "secret":
            return {}
        if _testing or _library_usage_mode:
            return await self._handle_testing_mode()
        yml_path = AsyncPath(settings_path / f"{self.adapter_name}.yml")
        await self._create_yml_file_if_needed(yml_path)
        yml_settings = await load.yaml(yml_path)
        if self.adapter_name == "debug":
            yml_settings = self._process_debug_settings(yml_settings)
        if not _deployed:
            await dump.yaml(yml_settings, yml_path, sort_keys=True)
        self._update_global_variables(yml_settings)
        return yml_settings or {}

    async def __call__(self) -> dict[str, t.Any]:
        data = {}
        if _testing or _library_usage_mode or Path.cwd().name != "acb":
            for field_name, field in (await self.load_yml_settings()).items():
                if field is not None:
                    data[field_name] = field
            return data
        return data


class SettingsProtocol(t.Protocol):
    model_config: t.ClassVar[SettingsConfigDict]

    def __init__(self, _secrets_path: AsyncPath = ..., **values: t.Any) -> None: ...

    _settings_build_values: t.Callable[..., t.Awaitable[dict[str, t.Any]]]
    settings_customize_sources: t.Callable[..., tuple[PydanticSettingsProtocol, ...]]


@rich.repr.auto
class Settings(BaseModel):
    model_config = SettingsConfigDict(
        extra="allow",
        arbitrary_types_allowed=True,
        validate_default=True,
        secrets_dir=Path(secrets_path),
        protected_namespaces=("model_", "settings_"),
    )

    def __init__(
        self, _secrets_path: AsyncPath = secrets_path, **values: t.Any
    ) -> None:
        build_settings_coro = self._settings_build_values(
            values, _secrets_path=_secrets_path
        )
        build_settings = asyncio.run(build_settings_coro)
        super().__init__(**build_settings)

    async def _settings_build_values(
        self, init_kwargs: dict[str, t.Any], _secrets_path: AsyncPath = secrets_path
    ) -> dict[str, t.Any]:
        init_settings: InitSettingsSource = InitSettingsSource(
            self.__class__, init_kwargs=init_kwargs
        )
        file_secret_settings: FileSecretSource = FileSecretSource(
            self.__class__, secrets_path=secrets_path
        )
        manager_secret_settings: ManagerSecretSource = ManagerSecretSource(
            self.__class__, secrets_path=secrets_path
        )
        yaml_settings: YamlSettingsSource = YamlSettingsSource(self.__class__)
        sources = self.settings_customize_sources(
            settings_cls=self.__class__,
            init_settings=t.cast(PydanticSettingsProtocol, init_settings),
            yaml_settings=t.cast(PydanticSettingsProtocol, yaml_settings),
            file_secret_settings=t.cast(PydanticSettingsProtocol, file_secret_settings),
            manager_secret_settings=t.cast(
                PydanticSettingsProtocol, manager_secret_settings
            ),
        )
        return deep_update(*reversed([await source() for source in sources]))

    @classmethod
    def settings_customize_sources(
        cls,
        settings_cls: type["Settings"],
        init_settings: PydanticSettingsProtocol,
        yaml_settings: PydanticSettingsProtocol,
        file_secret_settings: PydanticSettingsProtocol,
        manager_secret_settings: PydanticSettingsProtocol,
    ) -> tuple[PydanticSettingsProtocol, ...]:
        sources = [init_settings, yaml_settings, file_secret_settings]
        if get_adapter("secret") is not None:
            sources.append(manager_secret_settings)
        return tuple(sources)


class DebugSettings(Settings):
    production: bool = False
    secrets: bool = False
    logger: bool = False


class AppSettings(Settings):
    name: str = root_path.stem
    secret_key: SecretStr = SecretStr(token_urlsafe(32))
    secure_salt: SecretStr = SecretStr(str(token_bytes(32)))
    title: str | None = None
    domain: str | None = None
    platform: Platform | None = None
    project: str | None = None
    region: str | None = None
    timezone: str | None = "US/Pacific"
    version: str | None = Field(default_factory=get_version_default)

    def __init__(self, **values: t.Any) -> None:
        super().__init__(**values)
        self.title = self.title or titleize(self.name)

    @field_validator("name")
    @classmethod
    def cloud_compliant_app_name(cls, v: str) -> str:
        not_ok = [" ", "_", "."]
        _name = v
        for p in not_ok:
            _name = _name.replace(p, "-")
        for p in punctuation.replace("-", ""):
            _name = _name.replace(p, "")
        app_name = _name.strip("-").lower()
        if len(app_name) < 3:
            raise SystemExit("App name to short")
        elif len(app_name) > 63:
            raise SystemExit("App name to long")
        return app_name


class _LibraryDebugSettings:
    def __init__(self) -> None:
        self.production = False
        self.secrets = False
        self.logger = False


class _LibraryAppSettings:
    def __init__(self, name: str = "library_app") -> None:
        self.name = name
        self.secret_key = SecretStr(token_urlsafe(32))
        self.secure_salt = SecretStr(str(token_bytes(32)))
        self.title = titleize(name)
        self.domain = None
        self.platform = None
        self.project = None
        self.region = None
        self.timezone = "US/Pacific"
        self.version = "0.1.0"


class AdapterMeta(type):
    _instances: dict[t.Any, t.Any] = {}

    def __call__(cls, *args: t.Any, **kwargs: t.Any) -> t.Any:
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


@t.runtime_checkable
class ConfigProtocol(t.Protocol):
    deployed: bool
    root_path: AsyncPath
    secrets_path: AsyncPath
    settings_path: AsyncPath
    tmp_path: AsyncPath
    debug: DebugSettings | None
    app: AppSettings | None

    def init(self) -> None: ...


@rich.repr.auto
@dataclass(config=ConfigDict(arbitrary_types_allowed=True, extra="allow"))
class Config(metaclass=AdapterMeta):
    deployed: bool = _deployed
    root_path: AsyncPath = root_path
    secrets_path: AsyncPath = secrets_path
    settings_path: AsyncPath = settings_path
    tmp_path: AsyncPath = tmp_path
    _debug: DebugSettings | None = None
    _app: AppSettings | None = None
    _initialized: bool = False

    def init(self, force: bool = False) -> None:
        global _config_initialized
        if _library_usage_mode and not force and not _testing:
            return
        if self._initialized and not force:
            return
        self._debug = DebugSettings()
        self._app = AppSettings()
        self._initialized = True
        _config_initialized = True

    def ensure_initialized(self) -> None:
        if not self._initialized:
            if _library_usage_mode:
                self._debug = t.cast(DebugSettings, _LibraryDebugSettings())
                self._app = t.cast(AppSettings, _LibraryAppSettings())
                self._initialized = True
            else:
                self.init(force=True)

    @property
    def debug(self) -> DebugSettings | None:
        if self._debug is None and not self._initialized:
            self.ensure_initialized()
        return self._debug

    @property
    def app(self) -> AppSettings | None:
        if self._app is None and not self._initialized:
            self.ensure_initialized()
        return self._app

    def __getattr__(self, item: str) -> t.Any:
        if item in self.__dict__:
            return self.__dict__[item]
        if "." in item:
            parts = item.split(".")
            current_level = self
            for part in parts:
                if hasattr(current_level, part):
                    current_level = getattr(current_level, part)
            return current_level
        raise AttributeError(f"'Config' object has no attribute '{item}'")


depends.set(Config)

if _should_initialize_eagerly():
    depends.get(Config).init()

_ADAPTER_LOCKS: WeakKeyDictionary = WeakKeyDictionary()

Logger = None


@rich.repr.auto
class AdapterBase(metaclass=AdapterMeta):
    config: Config = depends()

    @property
    def logger(self):
        if not hasattr(self, "_logger"):
            try:
                Logger = import_adapter("logger")
                self._logger = depends.get(Logger)
            except Exception:
                import logging

                self._logger = logging.getLogger(self.__class__.__name__)
        return self._logger

    @logger.setter
    def logger(self, value: t.Any) -> None:
        self._logger = value

    @logger.deleter
    def logger(self) -> None:
        if hasattr(self, "_logger"):
            delattr(self, "_logger")

    def __init__(self, **kwargs: t.Any) -> None:
        self._client = None
        self._resource_cache: dict[str, t.Any] = {}
        self._initialization_args = kwargs
        if self not in _ADAPTER_LOCKS:
            _ADAPTER_LOCKS[self] = asyncio.Lock()

    async def _ensure_client(self) -> t.Any:
        if self._client is None:
            lock = _ADAPTER_LOCKS[self]
            async with lock:
                if self._client is None:
                    self._client = await self._create_client()
        return self._client

    async def _create_client(self) -> t.Any:
        raise NotImplementedError("Subclasses must implement _create_client()")

    async def _ensure_resource(
        self, resource_name: str, factory_func: t.Callable[[], t.Awaitable[t.Any]]
    ) -> t.Any:
        if resource_name not in self._resource_cache:
            lock = _ADAPTER_LOCKS[self]
            async with lock:
                if resource_name not in self._resource_cache:
                    self._resource_cache[resource_name] = await factory_func()
        return self._resource_cache[resource_name]

    async def _cleanup_resources(self) -> None:
        if self._client is not None:
            if hasattr(self._client, "close"):
                await self._client.close()
            elif hasattr(self._client, "aclose"):
                await self._client.aclose()
        for resource in self._resource_cache.values():
            if hasattr(resource, "close"):
                try:
                    await resource.close()
                except Exception as e:
                    self.logger.warning(f"Error closing resource: {e}")

    async def init(self) -> None:
        pass



"""Tests for the Sentry monitoring adapter."""

import typing as t
from unittest.mock import MagicMock, patch

import pytest
from acb.adapters.monitoring.sentry import Monitoring, MonitoringSettings
from acb.config import Config


@pytest.fixture
def mock_sentry_init() -> t.Generator[MagicMock]:
    mock = MagicMock()
    with patch("acb.adapters.monitoring.sentry.sentry_init", mock):
        yield mock


@pytest.fixture
def mock_integrations() -> dict[str, MagicMock]:
    integrations = {
        "GcpIntegration": MagicMock(),
        "AsyncioIntegration": MagicMock(),
    }

    integrations["GcpIntegration"].return_value = MagicMock(
        name="GcpIntegration_instance"
    )
    integrations["AsyncioIntegration"].return_value = MagicMock(
        name="AsyncioIntegration_instance"
    )

    return integrations


@pytest.fixture
def patch_integrations(
    mock_integrations: dict[str, MagicMock],
) -> t.Generator[None]:
    with (
        patch(
            "acb.adapters.monitoring.sentry.GcpIntegration",
            mock_integrations["GcpIntegration"],
        ),
        patch(
            "acb.adapters.monitoring.sentry.AsyncioIntegration",
            mock_integrations["AsyncioIntegration"],
        ),
    ):
        yield


@pytest.fixture
def monitoring_adapter(
    mock_sentry_init: MagicMock, patch_integrations: None
) -> Monitoring:
    mock_config = MagicMock()
    mock_config.monitoring = MonitoringSettings(
        sentry_dsn="https://test@sentry.io/123456",
        sample_rate=0.75,
        debug=True,
        traces_sample_rate=0.5,
        profiles_sample_rate=0.25,
    )
    mock_config.app.name = "test_app"
    mock_config.app.version = "1.0.0"
    mock_config.deployed = False

    adapter = Monitoring()
    adapter.config = mock_config

    return adapter


@pytest.mark.asyncio
async def test_init(
    monitoring_adapter: Monitoring,
    mock_sentry_init: MagicMock,
    mock_integrations: dict[str, MagicMock],
) -> None:
    await monitoring_adapter.init()

    mock_sentry_init.assert_called_once_with(
        dsn="https://test@sentry.io/123456",
        server_name="test_app",
        release="1.0.0",
        environment="development",
        sample_rate=0.75,
        debug=True,
        traces_sample_rate=0.5,
        profiles_sample_rate=0.25,
        integrations=[
            mock_integrations["GcpIntegration"].return_value,
            mock_integrations["AsyncioIntegration"].return_value,
        ],
    )


@pytest.mark.asyncio
async def test_init_production_environment(
    monitoring_adapter: Monitoring, mock_sentry_init: MagicMock
) -> None:
    monitoring_adapter.config.deployed = True

    await monitoring_adapter.init()

    called_args = mock_sentry_init.call_args[1]
    assert called_args["environment"] == "production"


class TestMonitoringSettings:
    def test_default_values(self) -> None:
        settings = MonitoringSettings()
        assert settings.sentry_dsn.get_secret_value() == "https://"
        assert settings.sample_rate == 1.0
        assert settings.debug is False
        assert settings.profiles_sample_rate == 0
        assert settings.traces_sample_rate == 0

    def test_valid_sample_rates(self) -> None:
        settings = MonitoringSettings(
            sample_rate=0.5, profiles_sample_rate=0.75, traces_sample_rate=0.25
        )
        assert settings.sample_rate == 0.5
        assert settings.profiles_sample_rate == 0.75
        assert settings.traces_sample_rate == 0.25

    def test_invalid_sample_rates(self) -> None:
        with pytest.raises(ValueError):
            MonitoringSettings(sample_rate=-0.1)

        with pytest.raises(ValueError):
            MonitoringSettings(sample_rate=1.1)

        with pytest.raises(ValueError):
            MonitoringSettings(profiles_sample_rate=-0.1)

        with pytest.raises(ValueError):
            MonitoringSettings(profiles_sample_rate=1.1)

    @patch("acb.adapters.monitoring.sentry.depends.inject")
    def test_init_with_config(self, mock_inject: MagicMock) -> None:
        mock_config = MagicMock(spec=Config)
        mock_config.deployed = False

        settings = MonitoringSettings()

        settings.__init__(config=mock_config)

        assert settings.sample_rate == 1.0

        mock_config.deployed = True
        settings.__init__(config=mock_config)

        assert settings.sample_rate == 1.0

    @patch("acb.adapters.monitoring.sentry.depends.inject")
    def test_init_with_explicit_sample_rate(self, mock_inject: MagicMock) -> None:
        mock_config = MagicMock(spec=Config)
        mock_config.deployed = True

        settings = MonitoringSettings(sample_rate=0.5)

        settings.__init__(config=mock_config, sample_rate=0.5)

        assert settings.sample_rate == 0.5



"""Tests for the Logfire monitoring adapter."""

import typing as t
from unittest.mock import MagicMock, PropertyMock, patch

import pytest
from pydantic import SecretStr
from acb.adapters.monitoring.logfire import Monitoring, MonitoringSettings
from acb.logger import Logger


@pytest.fixture
def mock_logfire_modules() -> t.Generator[dict[str, MagicMock]]:
    mocks = {
        "configure": MagicMock(),
        "instrument_pydantic": MagicMock(),
        "instrument_httpx": MagicMock(),
        "instrument_redis": MagicMock(),
        "instrument_sqlalchemy": MagicMock(),
        "instrument_system_metrics": MagicMock(),
        "loguru_handler": MagicMock(),
    }

    with (
        patch("acb.adapters.monitoring.logfire.configure", mocks["configure"]),
        patch(
            "acb.adapters.monitoring.logfire.instrument_pydantic",
            mocks["instrument_pydantic"],
        ),
        patch(
            "acb.adapters.monitoring.logfire.instrument_httpx",
            mocks["instrument_httpx"],
        ),
        patch(
            "acb.adapters.monitoring.logfire.instrument_redis",
            mocks["instrument_redis"],
        ),
        patch(
            "acb.adapters.monitoring.logfire.instrument_sqlalchemy",
            mocks["instrument_sqlalchemy"],
        ),
        patch(
            "acb.adapters.monitoring.logfire.instrument_system_metrics",
            mocks["instrument_system_metrics"],
        ),
        patch(
            "acb.adapters.monitoring.logfire.loguru_handler", mocks["loguru_handler"]
        ),
    ):
        yield mocks


@pytest.fixture
def mock_adapters() -> t.Generator[MagicMock]:
    mock = MagicMock()

    adapters = []
    for name in ("loguru", "httpx", "redis", "sqlalchemy", "other_adapter"):
        adapter_mock = MagicMock()
        type(adapter_mock).name = PropertyMock(return_value=name)
        adapters.append(adapter_mock)

    mock.return_value = adapters

    with patch("acb.adapters.monitoring.logfire.get_installed_adapters", mock):
        yield mock


@pytest.fixture
def mock_sql() -> t.Generator[MagicMock]:
    mock = MagicMock()
    mock.engine = MagicMock()

    with patch("acb.adapters.monitoring.logfire.depends.get", return_value=mock):
        yield mock


@pytest.fixture
def monitoring_adapter(
    mock_logfire_modules: dict[str, MagicMock],
    mock_adapters: MagicMock,
    mock_sql: MagicMock,
) -> Monitoring:
    mock_config = MagicMock()
    mock_config.monitoring = MonitoringSettings(logfire_token=SecretStr("test_token"))
    mock_config.app.name = "test_app"
    mock_config.app.version = "1.0.0"

    mock_logger = MagicMock(spec=Logger)

    adapter = Monitoring()
    adapter.config = mock_config
    adapter.logger = mock_logger

    return adapter


@pytest.mark.asyncio
async def test_init(
    monitoring_adapter: Monitoring,
    mock_logfire_modules: dict[str, MagicMock],
    mock_sql: MagicMock,
) -> None:
    with patch("loguru.logger") as mock_loguru_logger:
        await monitoring_adapter.init()

        mock_logfire_modules["configure"].assert_called_once_with(
            token="test_token",
            service_name="test_app",
            service_version="1.0.0",
        )

        mock_logfire_modules["instrument_pydantic"].assert_called_once_with(
            record="all"
        )

        mock_logfire_modules["loguru_handler"].assert_called_once()
        mock_loguru_logger.configure.assert_called_once()
        mock_logfire_modules["instrument_httpx"].assert_called_once()
        mock_logfire_modules["instrument_redis"].assert_called_once()
        mock_logfire_modules["instrument_sqlalchemy"].assert_called_once_with(
            engine=mock_sql.engine
        )
        mock_logfire_modules["instrument_system_metrics"].assert_called_once()


@pytest.mark.asyncio
async def test_init_with_no_adapters(
    monitoring_adapter: Monitoring, mock_logfire_modules: dict[str, MagicMock]
) -> None:
    with patch(
        "acb.adapters.monitoring.logfire.get_installed_adapters", return_value=[]
    ):
        await monitoring_adapter.init()

    mock_logfire_modules["configure"].assert_called_once()
    mock_logfire_modules["instrument_pydantic"].assert_called_once()
    mock_logfire_modules["instrument_system_metrics"].assert_called_once()

    # configure is called once during init, not during adapter-specific setup
    mock_logfire_modules["instrument_httpx"].assert_not_called()
    mock_logfire_modules["instrument_redis"].assert_not_called()
    mock_logfire_modules["instrument_sqlalchemy"].assert_not_called()


class TestMonitoringSettings:
    def test_valid_traces_sample_rate(self) -> None:
        settings = MonitoringSettings(traces_sample_rate=0.5)
        assert settings.traces_sample_rate == 0.5

        settings = MonitoringSettings(traces_sample_rate=0.0)
        assert settings.traces_sample_rate == 0.0

        settings = MonitoringSettings(traces_sample_rate=1.0)
        assert settings.traces_sample_rate == 1.0

    def test_invalid_traces_sample_rate(self) -> None:
        with pytest.raises(ValueError):
            MonitoringSettings(traces_sample_rate=-0.1)

        with pytest.raises(ValueError):
            MonitoringSettings(traces_sample_rate=1.1)



import typing as t

from aiocache import BaseCache
from aiocache.serializers import BaseSerializer
from msgspec import msgpack
from acb.actions.compress import compress, decompress
from acb.config import Config, Settings
from acb.depends import depends
from acb.logger import Logger


class CacheBaseSettings(Settings):
    default_ttl: int = 86400
    query_ttl: int = 600
    response_ttl: int | None = 3600
    template_ttl: int | None = 86400

    @depends.inject
    def __init__(self, config: Config = depends(), **values: t.Any) -> None:
        super().__init__(**values)
        self.response_ttl = self.default_ttl if config.deployed else 1


class MsgPackSerializer(BaseSerializer):
    def __init__(self, *args: t.Any, use_list: bool = True, **kwargs: t.Any) -> None:
        self.use_list = use_list
        super().__init__(*args, **kwargs)

    def dumps(self, value: t.Any) -> str:
        msgpack_data = msgpack.encode(value)
        return compress.brotli(msgpack_data).decode("latin-1")

    def loads(self, value: str) -> t.Any:
        if not value:
            return None
        data_bytes = value.encode("latin-1")
        msgpack_data = decompress.brotli(data_bytes)
        if not msgpack_data:
            return None
        try:
            msgpack_bytes = msgpack_data.encode("latin-1")
        except AttributeError:
            msgpack_bytes = t.cast(bytes, msgpack_data)
        return msgpack.decode(msgpack_bytes)


class CacheProtocol(t.Protocol):
    async def set(self, key: str, value: bytes, ttl: int | None = None) -> None: ...

    async def get(self, key: str) -> bytes | None: ...

    async def exists(self, key: str) -> bool: ...

    async def clear(self, namespace: str) -> None: ...

    async def scan(self, pattern: str) -> t.AsyncIterator[str]: ...


class CacheBase(BaseCache):
    config: Config = depends()
    logger: Logger = depends()

    def __init__(self, **kwargs: t.Any) -> None:
        super().__init__()
        self._client = None
        self._client_lock = None

    async def _ensure_client(self) -> t.Any:
        if self._client is None:
            if self._client_lock is None:
                import asyncio

                self._client_lock = asyncio.Lock()
            async with self._client_lock:
                if self._client is None:
                    self._client = await self._create_client()
        return self._client

    async def _create_client(self) -> t.Any:
        raise NotImplementedError("Subclasses must implement _create_client()")



"""Tests for the SecretManager adapter."""

import typing as t
from unittest.mock import AsyncMock, MagicMock, PropertyMock, patch

import pytest
from google.api_core.exceptions import AlreadyExists, PermissionDenied
from acb.adapters.secret.secret_manager import Secret, SecretSettings
from acb.config import Config
from acb.logger import Logger


class AsyncIterator:
    def __init__(self, items: list[t.Any]) -> None:
        self.items = items.copy()

    def __aiter__(self) -> t.Self:
        return self

    async def __anext__(self) -> t.Any:
        if not self.items:
            raise StopAsyncIteration
        return self.items.pop(0)


class TestSecretManager:
    @pytest.fixture
    def mock_config(self) -> MagicMock:
        mock_config = MagicMock(spec=Config)
        mock_secret_settings = MagicMock(spec=SecretSettings)
        mock_secret_settings.project = "test-project"

        mock_config.secret = mock_secret_settings
        return mock_config

    @pytest.fixture
    def mock_logger(self) -> MagicMock:
        return MagicMock(spec=Logger)

    @pytest.fixture
    def mock_client(self) -> MagicMock:
        client = MagicMock()
        client.secret_path = MagicMock(
            return_value="projects/test-project/secrets/acb_test_secret"
        )
        client.access_secret_version = AsyncMock()
        client.create_secret = AsyncMock()
        client.add_secret_version = AsyncMock()
        client.delete_secret = AsyncMock()
        client.list_secret_versions = AsyncMock()
        client.list_secrets = AsyncMock()

        mock_response = MagicMock()
        mock_payload = MagicMock()
        mock_payload.data = b"test-value"
        mock_response.payload = mock_payload
        client.access_secret_version.return_value = mock_response

        mock_secret = MagicMock()
        mock_secret.name = "projects/test-project/secrets/acb_new_secret"
        client.create_secret.return_value = mock_secret

        return client

    @pytest.fixture
    def secret_manager(
        self, mock_config: MagicMock, mock_logger: MagicMock, mock_client: MagicMock
    ) -> t.Generator[Secret]:
        adapter = Secret()
        adapter.config = mock_config
        adapter.logger = mock_logger
        adapter.prefix = "acb_"
        adapter.project = "test-project"
        adapter.parent = "projects/test-project"

        with patch.object(
            Secret, "client", new_callable=PropertyMock
        ) as mock_client_prop:
            mock_client_prop.return_value = mock_client
            yield adapter

    def test_extract_secret_name(self, secret_manager: Secret) -> None:
        secret_path = "projects/test-project/secrets/acb_app_settings"
        result = secret_manager.extract_secret_name(secret_path)
        assert result == "app_settings"

    @pytest.mark.asyncio
    async def test_list(self, secret_manager: Secret, mock_client: MagicMock) -> None:
        mock_secret1 = MagicMock()
        mock_secret1.name = "projects/test-project/secrets/acb_test_secret1"
        mock_secret2 = MagicMock()
        mock_secret2.name = "projects/test-project/secrets/acb_adapter_secret2"

        secrets = [mock_secret1, mock_secret2]
        mock_client.list_secrets.return_value = AsyncIterator(secrets)

        result = await secret_manager.list()

        assert mock_client.list_secrets.call_count == 1
        mock_client.list_secrets.assert_called_once()

        assert len(result) == 2
        assert "test_secret1" in result
        assert "adapter_secret2" in result

        mock_client.list_secrets.reset_mock()

        adapter_secret = MagicMock()
        adapter_secret.name = "projects/test-project/secrets/acb_adapter_secret2"
        mock_client.list_secrets.return_value = AsyncIterator([adapter_secret])

        result = await secret_manager.list(adapter="adapter")

        assert mock_client.list_secrets.call_count == 1
        assert len(result) == 1
        assert "adapter_secret2" in result

        mock_client.list_secrets.reset_mock()
        mock_client.list_secrets.side_effect = PermissionDenied("Permission denied")

        with pytest.raises(SystemExit):
            await secret_manager.list()

    @pytest.mark.asyncio
    async def test_get(self, secret_manager: Secret, mock_client: MagicMock) -> None:
        name = "test_secret"

        result = await secret_manager.get(name)

        mock_client.access_secret_version.assert_called_once()

        assert result == "test-value"

        mock_client.access_secret_version.reset_mock()

        await secret_manager.get(name, version="2")

        assert mock_client.access_secret_version.call_count == 1

    @pytest.mark.asyncio
    async def test_create(self, secret_manager: Secret, mock_client: MagicMock) -> None:
        name = "new_secret"
        value = "secret-value"

        await secret_manager.create(name, value)

        mock_client.create_secret.assert_called_once()
        mock_client.add_secret_version.assert_called_once()

        mock_client.create_secret.reset_mock()
        mock_client.add_secret_version.reset_mock()

        mock_client.create_secret.side_effect = AlreadyExists("Secret already exists")

        await secret_manager.create(name, value)

        mock_client.create_secret.assert_called_once()

        mock_client.add_secret_version.assert_not_called()

    @pytest.mark.asyncio
    async def test_update(self, secret_manager: Secret, mock_client: MagicMock) -> None:
        name = "existing_secret"
        value = "updated-value"

        await secret_manager.update(name, value)

        mock_client.secret_path.assert_called_once_with("test-project", name)
        mock_client.add_secret_version.assert_called_once()

    @pytest.mark.asyncio
    async def test_set(self, secret_manager: Secret) -> None:
        with (
            patch.object(
                secret_manager, "exists", new_callable=AsyncMock
            ) as mock_exists,
            patch.object(
                secret_manager, "update", new_callable=AsyncMock
            ) as mock_update,
            patch.object(
                secret_manager, "create", new_callable=AsyncMock
            ) as mock_create,
        ):
            mock_exists.return_value = True

            await secret_manager.set("existing_secret", "updated-value")

            mock_update.assert_called_once_with("existing_secret", "updated-value")
            mock_create.assert_not_called()

            mock_exists.reset_mock()
            mock_update.reset_mock()
            mock_exists.return_value = False

            await secret_manager.set("new_secret", "new-value")

            mock_create.assert_called_once_with("new_secret", "new-value")
            mock_update.assert_not_called()

    @pytest.mark.asyncio
    async def test_exists(self, secret_manager: Secret) -> None:
        with patch.object(secret_manager, "get", new_callable=AsyncMock) as mock_get:
            mock_get.return_value = "test-value"

            result = await secret_manager.exists("existing_secret")

            assert result
            mock_get.assert_called_once_with("existing_secret")

            mock_get.reset_mock()
            mock_get.side_effect = Exception("Not found")

            result = await secret_manager.exists("non_existing_secret")

            assert not result

    @pytest.mark.asyncio
    async def test_delete(self, secret_manager: Secret, mock_client: MagicMock) -> None:
        name = "delete_secret"

        await secret_manager.delete(name)

        mock_client.secret_path.assert_called_once_with("test-project", name)
        mock_client.delete_secret.assert_called_once()

    @pytest.mark.asyncio
    async def test_list_versions(
        self, secret_manager: Secret, mock_client: MagicMock
    ) -> None:
        name = "test_secret"

        mock_version = MagicMock()
        mock_version.name = "projects/test-project/secrets/acb_test_secret/versions/1"

        mock_client.list_secret_versions.return_value = AsyncIterator([mock_version])

        result = await secret_manager.list_versions(name)

        mock_client.secret_path.assert_called_once_with("test-project", name)
        mock_client.list_secret_versions.assert_called_once()

        assert len(result) == 1
        assert result[0] == "1"



import asyncio
import typing as t
from contextvars import ContextVar
from inspect import currentframe

import nest_asyncio
import rich.repr
from anyio import Path as AsyncPath
from pydantic import BaseModel, ConfigDict
from rich import box
from rich.console import RenderableType
from rich.padding import Padding
from rich.table import Table

from .actions import Action, action_registry, register_actions
from .adapters import Adapter, _deployed, get_adapters, register_adapters
from .console import console

nest_asyncio.apply()


@rich.repr.auto
class Pkg(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    name: str
    path: AsyncPath
    actions: list[Action] = []
    adapters: list[Adapter] = []


pkg_registry: ContextVar[list[Pkg]] = ContextVar("pkg_registry", default=[])
_lazy_registration_queue: list[tuple[str, AsyncPath]] = []
_registration_completed: bool = False


def register_pkg() -> None:
    frame = currentframe()
    if frame is not None and frame.f_back is not None:
        path = AsyncPath(frame.f_back.f_code.co_filename).parent
    else:
        raise RuntimeError("Could not determine caller frame")
    name = path.stem
    if name not in [item[0] for item in _lazy_registration_queue]:
        _lazy_registration_queue.append((name, path))


async def _process_registration_queue() -> None:
    global _registration_completed
    if _registration_completed or not _lazy_registration_queue:
        return
    registry = pkg_registry.get()
    existing_names = {p.name for p in registry}
    for name, path in _lazy_registration_queue:
        if name not in existing_names:
            actions = await register_actions(path)
            adapters = await register_adapters(path)
            pkg = Pkg(name=name, path=path, actions=actions, adapters=adapters)
            registry.append(pkg)
            existing_names.add(name)
    _lazy_registration_queue.clear()
    _registration_completed = True


def ensure_registration() -> None:
    if _lazy_registration_queue and not _registration_completed:
        try:
            asyncio.get_running_loop()
            asyncio.create_task(_process_registration_queue())
        except RuntimeError:
            asyncio.run(_process_registration_queue())


try:
    from pathlib import Path

    from anyio import Path as AsyncPath

    from .adapters import Adapter, _update_adapter_caches, adapter_registry

    if not any(
        a.pkg == "acb" and a.category == "models" for a in adapter_registry.get()
    ):
        acb_adapters_path = AsyncPath(Path(__file__).parent / "adapters")
        essential_adapters = [
            Adapter(
                name="sqlmodel",
                class_name="Models",
                category="models",
                pkg="acb",
                module="acb.adapters.models.sqlmodel",
                enabled=False,
                installed=False,
                path=acb_adapters_path / "models" / "sqlmodel.py",
            ),
            Adapter(
                name="memory",
                class_name="Cache",
                category="cache",
                pkg="acb",
                module="acb.adapters.cache.memory",
                enabled=False,
                installed=False,
                path=acb_adapters_path / "cache" / "memory.py",
            ),
            Adapter(
                name="file",
                class_name="Storage",
                category="storage",
                pkg="acb",
                module="acb.adapters.storage.file",
                enabled=False,
                installed=False,
                path=acb_adapters_path / "storage" / "file.py",
            ),
        ]
        registry = adapter_registry.get()
        registry.extend(essential_adapters)
        _update_adapter_caches()
except Exception as e:
    print(f"Warning: ACB essential adapters registration failed: {e}")
    pass
table_args: dict[str, t.Any] = {
    "show_lines": True,
    "box": box.ROUNDED,
    "min_width": 100,
    "border_style": "bold blue",
}


def display_components() -> None:
    ensure_registration()
    if not _deployed:
        pkgs = Table(
            title="[b][u bright_white]A[/u bright_white][bright_green]synchronous [u bright_white]C[/u bright_white][bright_green]omponent [u bright_white]B[/u bright_white][bright_green]ase[/b]",
            **table_args,
        )
        for prop in ("Pkg", "Path"):
            pkgs.add_column(prop)
        for i, pkg in enumerate(pkg_registry.get()):
            pkgs.add_row(
                pkg.name,
                str(pkg.path),
                style="bold white" if i % 2 else "bold white on blue",
            )
        pkgs_padded: RenderableType = Padding(pkgs, (2, 4, 0, 4))
        console.print(pkgs_padded)
        actns = Table(
            title="[b][u bright_white]A[/u bright_white][bright_green]ctions",
            **table_args,
        )
        for prop in ("Name", "Pkg", "Methods"):
            actns.add_column(prop)
        for i, action in enumerate(action_registry.get()):
            actns.add_row(
                action.name,
                action.pkg,
                ", ".join(action.methods),
                style="bold white" if i % 2 else "bold white on blue",
            )
        actns_padded: RenderableType = Padding(actns, (0, 4, 1, 4))
        console.print(actns_padded)
        adptrs = Table(
            title="[b][u bright_white]A[/u bright_white][bright_green]datpters",
            **table_args,
        )
        for prop in ("Category", "Name", "Pkg"):
            adptrs.add_column(prop)
        for i, adapter in enumerate(get_adapters()):
            adptrs.add_row(
                adapter.category,
                adapter.name,
                adapter.pkg,
                style="bold white" if i % 2 else "bold white on blue",
            )
        adptrs_padded: RenderableType = Padding(adptrs, (0, 4, 1, 4))
        console.print(adptrs_padded)



"""Tests for the Infisical adapter."""

import typing as t
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from acb.adapters.secret.infisical import Secret, SecretSettings
from acb.config import Config
from acb.logger import Logger


class MockClient(MagicMock):
    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self.secrets = MagicMock()
        self.auth = MagicMock()
        self.auth.universal_auth = MagicMock()


class TestInfisical:
    @pytest.fixture
    def mock_config(self) -> MagicMock:
        mock_config = MagicMock(spec=Config)
        mock_secret_settings = MagicMock(spec=SecretSettings)
        mock_secret_settings.host = "https://app.infisical.com"
        mock_secret_settings.client_id = "test-client-id"
        mock_secret_settings.client_secret = "test-client-secret"
        mock_secret_settings.token = "test-token"
        mock_secret_settings.project_id = "test-project-id"
        mock_secret_settings.environment = "dev"
        mock_secret_settings.secret_path = "/"
        mock_secret_settings.cache_ttl = 60

        mock_config.secret = mock_secret_settings
        return mock_config

    @pytest.fixture
    def mock_logger(self) -> MagicMock:
        return MagicMock(spec=Logger)

    @pytest.fixture
    def mock_infisical_client(self, infisical_adapter: Secret) -> MagicMock:
        client = MockClient()

        prefix = infisical_adapter.prefix

        mock_secret_response = MagicMock()
        mock_secret_response.secretValue = "test-value"
        client.secrets.get_secret_by_name = MagicMock(return_value=mock_secret_response)

        mock_list_response = MagicMock()
        mock_secret1 = MagicMock()
        mock_secret1.secretKey = f"{prefix}test_secret1"
        mock_secret2 = MagicMock()
        mock_secret2.secretKey = f"{prefix}adapter_secret2"
        mock_list_response.secrets = [mock_secret1, mock_secret2]
        client.secrets.list_secrets = MagicMock(return_value=mock_list_response)

        client.secrets.create_secret_by_name = MagicMock(return_value=None)
        client.secrets.update_secret_by_name = MagicMock(return_value=None)
        client.secrets.delete_secret_by_name = MagicMock(return_value=None)

        return client

    @pytest.fixture
    def infisical_adapter(
        self, mock_config: MagicMock, mock_logger: MagicMock
    ) -> Secret:
        adapter = Secret()
        adapter.config = mock_config
        adapter.logger = mock_logger
        return adapter

    @pytest.mark.asyncio
    async def test_extract_secret_name(self, infisical_adapter: Secret) -> None:
        prefix = infisical_adapter.prefix

        secret_path = f"{prefix}app_settings"
        result = infisical_adapter.extract_secret_name(secret_path)
        assert result == "app_settings"

    @pytest.mark.asyncio
    async def test_list(
        self, infisical_adapter: Secret, mock_infisical_client: MagicMock
    ) -> None:
        infisical_adapter._client = mock_infisical_client

        async def async_mock_func(*args: t.Any, **kwargs: t.Any) -> t.Any:
            func = args[0]
            args = args[1:]
            return func(*args, **kwargs)

        with patch("asyncio.to_thread", async_mock_func):
            result = await infisical_adapter.list()
            assert len(result) == 2
            assert "test_secret1" in result
            assert "adapter_secret2" in result

            prefix = infisical_adapter.prefix

            mock_list_response2 = MagicMock()
            mock_secret3 = MagicMock()
            mock_secret3.secretKey = f"{prefix}adapter_secret3"
            mock_list_response2.secrets = [mock_secret3]

            mock_infisical_client.secrets.list_secrets.reset_mock()
            mock_infisical_client.secrets.list_secrets.return_value = (
                mock_list_response2
            )

            result = await infisical_adapter.list(adapter="adapter")
            assert len(result) == 1
            assert "adapter_secret3" in result

    @pytest.mark.asyncio
    async def test_get(
        self, infisical_adapter: Secret, mock_infisical_client: MagicMock
    ) -> None:
        infisical_adapter._client = mock_infisical_client

        expected_prefix = infisical_adapter.prefix

        async def async_mock_func(*args: t.Any, **kwargs: t.Any) -> t.Any:
            func = args[0]
            args = args[1:]
            return func(*args, **kwargs)

        with patch("asyncio.to_thread", new=async_mock_func):
            result = await infisical_adapter.get("test_secret")

            mock_infisical_client.secrets.get_secret_by_name.assert_called_once()
            call_kwargs = (
                mock_infisical_client.secrets.get_secret_by_name.call_args.kwargs
            )
            assert call_kwargs["secret_name"] == f"{expected_prefix}test_secret"
            assert call_kwargs["project_id"] == "test-project-id"
            assert call_kwargs["environment_slug"] == "dev"

            assert result == "test-value"

            await infisical_adapter.get("test_secret", version="v1")
            call_kwargs = (
                mock_infisical_client.secrets.get_secret_by_name.call_args.kwargs
            )
            assert call_kwargs["version"] == "v1"

    @pytest.mark.asyncio
    async def test_create(
        self, infisical_adapter: Secret, mock_infisical_client: MagicMock
    ) -> None:
        infisical_adapter._client = mock_infisical_client

        expected_prefix = infisical_adapter.prefix

        async def async_mock_func(*args: t.Any, **kwargs: t.Any) -> t.Any:
            func = args[0]
            args = args[1:]
            return func(*args, **kwargs)

        with patch("asyncio.to_thread", new=async_mock_func):
            await infisical_adapter.create("new_secret", "secret-value")

            mock_infisical_client.secrets.create_secret_by_name.assert_called_once()
            call_kwargs = (
                mock_infisical_client.secrets.create_secret_by_name.call_args.kwargs
            )
            assert call_kwargs["secret_name"] == f"{expected_prefix}new_secret"
            assert call_kwargs["project_id"] == "test-project-id"
            assert call_kwargs["secret_value"] == "secret-value"

    @pytest.mark.asyncio
    async def test_update(
        self, infisical_adapter: Secret, mock_infisical_client: MagicMock
    ) -> None:
        infisical_adapter._client = mock_infisical_client

        expected_prefix = infisical_adapter.prefix

        async def async_mock_func(*args: t.Any, **kwargs: t.Any) -> t.Any:
            func = args[0]
            args = args[1:]
            return func(*args, **kwargs)

        with patch("asyncio.to_thread", new=async_mock_func):
            await infisical_adapter.update("update_secret", "updated-value")

            mock_infisical_client.secrets.update_secret_by_name.assert_called_once()
            call_kwargs = (
                mock_infisical_client.secrets.update_secret_by_name.call_args.kwargs
            )
            assert (
                call_kwargs["current_secret_name"] == f"{expected_prefix}update_secret"
            )
            assert call_kwargs["project_id"] == "test-project-id"
            assert call_kwargs["secret_value"] == "updated-value"

    @pytest.mark.asyncio
    async def test_delete(
        self, infisical_adapter: Secret, mock_infisical_client: MagicMock
    ) -> None:
        infisical_adapter._client = mock_infisical_client

        expected_prefix = infisical_adapter.prefix

        async def async_mock_func(*args: t.Any, **kwargs: t.Any) -> t.Any:
            func = args[0]
            args = args[1:]
            return func(*args, **kwargs)

        with patch("asyncio.to_thread", async_mock_func):
            await infisical_adapter.delete("delete_secret")

            mock_infisical_client.secrets.delete_secret_by_name.assert_called_once()
            call_kwargs = (
                mock_infisical_client.secrets.delete_secret_by_name.call_args.kwargs
            )
            assert call_kwargs["secret_name"] == f"{expected_prefix}delete_secret"
            assert call_kwargs["project_id"] == "test-project-id"

    @pytest.mark.asyncio
    async def test_exists(self, infisical_adapter: Secret) -> None:
        with patch.object(infisical_adapter, "get") as mock_get:
            mock_get.return_value = "test-value"
            result = await infisical_adapter.exists("existing_secret")
            assert result
            mock_get.assert_called_once_with("existing_secret")

            mock_get.reset_mock()
            mock_get.side_effect = Exception("Not found")
            result = await infisical_adapter.exists("non_existing_secret")
            assert not result

    @pytest.mark.asyncio
    async def test_set(self, infisical_adapter: Secret) -> None:
        with (
            patch.object(infisical_adapter, "exists") as mock_exists,
            patch.object(infisical_adapter, "update") as mock_update,
            patch.object(infisical_adapter, "create") as mock_create,
        ):
            mock_exists.return_value = True
            await infisical_adapter.set("existing_secret", "updated-value")
            mock_update.assert_called_once_with("existing_secret", "updated-value")
            mock_create.assert_not_called()

            mock_exists.reset_mock()
            mock_update.reset_mock()
            mock_exists.return_value = False

            await infisical_adapter.set("new_secret", "new-value")
            mock_create.assert_called_once_with("new_secret", "new-value")
            mock_update.assert_not_called()

    @pytest.mark.asyncio
    async def test_list_versions(self, infisical_adapter: Secret) -> None:
        with patch.object(infisical_adapter, "logger") as mock_logger:
            result = await infisical_adapter.list_versions("test_secret")
            assert not result
            mock_logger.warning.assert_called_once()

    @pytest.mark.asyncio
    async def test_init(
        self,
        infisical_adapter: Secret,
        mock_infisical_client: MagicMock,
        mock_logger: MagicMock,
    ) -> None:
        infisical_adapter._client = mock_infisical_client
        infisical_adapter.logger = mock_logger

        async def async_mock_func(*args: t.Any, **kwargs: t.Any) -> t.Any:
            func = args[0]
            args = args[1:]
            return func(*args, **kwargs)

        mock_list = AsyncMock(return_value=["secret1", "secret2"])

        with patch.object(infisical_adapter, "list", mock_list):
            with patch("asyncio.to_thread", async_mock_func):
                await infisical_adapter.init(logger=mock_logger)

                mock_list.assert_called_once()
                mock_logger.info.assert_called_once_with(
                    "Infisical secret adapter initialized successfully"
                )

        mock_logger.reset_mock()
        mock_list.reset_mock()

        mock_list = AsyncMock(side_effect=Exception("Test error"))
        with patch.object(infisical_adapter, "list", mock_list):
            with patch("asyncio.to_thread", async_mock_func):
                with pytest.raises(Exception):
                    await infisical_adapter.init(logger=mock_logger)

                mock_logger.error.assert_called_once()



import typing as t

from aiocache.backends.redis import RedisBackend
from aiocache.serializers import PickleSerializer
from coredis.cache import TrackingCache
from coredis.client import Redis, RedisCluster
from pydantic import SecretStr
from acb.config import Config
from acb.debug import debug
from acb.depends import depends

from ._base import CacheBase, CacheBaseSettings


class CacheSettings(CacheBaseSettings):
    host: SecretStr = SecretStr("127.0.0.1")
    local_host: str = "127.0.0.1"
    port: int | None = 6379
    cluster: bool | None = False
    connect_timeout: float | None = 3
    max_connections: int | None = 50
    health_check_interval: int | None = 0
    retry_on_timeout: bool | None = True

    @depends.inject
    def __init__(self, config: Config = depends(), **values: t.Any) -> None:
        super().__init__(**values)
        self.host = SecretStr(self.local_host) if not config.deployed else self.host


class Cache(CacheBase, RedisBackend):
    def __init__(self, redis_url: str | None = None, **kwargs: t.Any) -> None:
        self.redis_url = redis_url
        self._init_kwargs = {k: v for k, v in kwargs.items() if k != "redis_url"}
        super().__init__()

    async def _create_client(self) -> Redis[t.Any] | RedisCluster[t.Any]:
        redis_kwargs = self._init_kwargs | {
            "host": self.config.cache.host.get_secret_value(),
            "port": self.config.cache.port,
            "client_name": self.config.app.name,
            "cache": TrackingCache(),
            "decode_responses": False,
            "connect_timeout": self.config.cache.connect_timeout,
            "max_connections": self.config.cache.max_connections,
            "health_check_interval": self.config.cache.health_check_interval,
            "retry_on_timeout": self.config.cache.retry_on_timeout,
        }
        if self.config.cache.cluster:
            self.logger.info("RedisCluster mode enabled")
            del redis_kwargs["health_check_interval"]
            return RedisCluster(**redis_kwargs)

        return Redis(**redis_kwargs)

    async def get_client(self) -> Redis[t.Any] | RedisCluster[t.Any]:
        return await self._ensure_client()

    async def _close(self, *args: t.Any, _conn: t.Any = None, **kwargs: t.Any) -> None:
        if self._client is not None:
            await self._client.close()

    async def _clear(
        self, namespace: str | None = None, _conn: t.Any = None
    ) -> t.Literal[True]:
        if not namespace:
            pattern = f"{self.config.app.name}:*"
        else:
            pattern = f"{self.config.app.name}:{namespace}:*"
        client = await self.get_client()
        keys = await client.keys(pattern)
        if keys:
            debug(keys)
            for key in keys:
                await client.unlink((key,))
        return True

    async def _exists(self, key: str, _conn: t.Any = None) -> bool:
        client = await self.get_client()
        number = await client.exists([key])
        return bool(number)

    async def init(self, *args: t.Any, **kwargs: t.Any) -> None:
        self._init_kwargs.update(kwargs)
        if not hasattr(self, "_namespace"):
            self._namespace = f"{self.config.app.name}:"
        if not hasattr(self, "_serializer"):
            self._serializer = PickleSerializer()


depends.set(Cache)



from collections.abc import Generator
from pathlib import Path
from typing import Any, Final
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from rich.console import Console
from acb.config import Config
from acb.debug import (
    colorized_stderr_print,
    debug,
    init_debug,
    pprint,
    print_debug_info,
)
from acb.logger import Logger

TEST_MODULE_NAME: Final[str] = "test_module"
TEST_DEBUG_MSG: Final[str] = "test debug message"


@pytest.fixture
def mock_console() -> MagicMock:
    console = MagicMock(spec=Console)
    console.print = MagicMock()
    return console


@pytest.fixture
def mock_logger() -> MagicMock:
    logger = MagicMock(spec=Logger)
    logger.debug = MagicMock()
    return logger


class TestColorizedStderrPrint:
    def test_colorized_stderr_print_with_asyncio(self) -> None:
        with patch("acb.debug.colorize") as mock_colorize:
            mock_colorize.return_value = "Colored test message"

            with patch("acb.debug.supportTerminalColorsInWindows") as mock_support:
                mock_support.return_value.__enter__ = MagicMock()
                mock_support.return_value.__exit__ = MagicMock()

                mock_coro: AsyncMock = AsyncMock()
                with patch("acb.debug.aprint", return_value=mock_coro) as mock_aprint:

                    def mock_run_impl(coro: Any) -> None:
                        return None

                    with patch(
                        "acb.debug.asyncio.run", side_effect=mock_run_impl
                    ) as mock_run:
                        colorized_stderr_print("Test message")

                        mock_colorize.assert_called_once_with("Test message")

                        mock_support.assert_called_once()

                        mock_aprint.assert_called_once_with(
                            "Colored test message", use_stderr=True
                        )

                        mock_run.assert_called_once()

    def test_colorized_stderr_print_with_print(self) -> None:
        with patch("acb.debug.colorize") as mock_colorize:
            mock_colorize.return_value = "Colored test message"

            with patch("acb.debug.supportTerminalColorsInWindows") as mock_support:
                mock_support.return_value.__enter__ = MagicMock()
                mock_support.return_value.__exit__ = MagicMock()

                with patch(
                    "acb.debug.asyncio.run", side_effect=Exception("Test exception")
                ):
                    with patch("sys.stderr") as mock_stderr:
                        with patch("builtins.print") as mock_print:
                            colorized_stderr_print("Test message")

                            mock_print.assert_called_once_with(
                                "Colored test message", file=mock_stderr
                            )

    def test_colorized_stderr_print_with_no_color_support(self) -> None:
        with patch("acb.debug.colorize", side_effect=ImportError):
            with patch("builtins.print") as mock_print:
                with patch("sys.stderr") as mock_stderr:
                    colorized_stderr_print("Test message")
                    mock_print.assert_called_once_with("Test message", file=mock_stderr)


class TestPrintDebugInfo:
    def test_print_debug_info_with_module_not_deployed(self) -> None:
        mock_module: MagicMock = MagicMock(spec=Path)
        with patch("acb.debug.get_calling_module", return_value=mock_module):
            with patch("acb.debug._deployed", False):
                with patch("acb.debug.colorized_stderr_print") as mock_print:
                    result: Any = print_debug_info("Test message")

                    mock_print.assert_called_once_with("Test message")

                    assert result is None

    def test_print_debug_info_with_module_deployed(self) -> None:
        mock_module: MagicMock = MagicMock(spec=Path)
        with patch("acb.debug.get_calling_module", return_value=mock_module):
            with patch("acb.debug._deployed", True):
                with patch("acb.debug.patch_record") as mock_patch:
                    result: Any = print_debug_info("Test message")

                    mock_patch.assert_called_once_with(mock_module, "Test message")

                    assert result is None

    def test_print_debug_info_without_module(self) -> None:
        with patch("acb.debug.get_calling_module", return_value=None):
            result: Any = print_debug_info("Test message")

            assert result is None


class TestPprint:
    @pytest.mark.asyncio
    async def test_pprint(self) -> None:
        with patch("acb.debug.pformat") as mock_pformat:
            mock_pformat.return_value = "Formatted object"

            with patch("acb.debug.aprint") as mock_aprint:
                await pprint({"test": "object"})

                mock_pformat.assert_called_once_with({"test": "object"})

                mock_aprint.assert_called_once_with("Formatted object", use_stderr=True)

    @pytest.mark.asyncio
    async def test_pprint_with_custom_width(self) -> None:
        test_data = {"key": "value" * 20}
        with patch("acb.debug.pformat") as mock_pformat:
            with patch("acb.debug.aprint") as mock_aprint:
                await pprint(test_data)
                mock_pformat.assert_called_once_with(test_data)
                mock_aprint.assert_called_once()


class TestInitDebug:
    @pytest.fixture
    def mock_config(self) -> Generator[MagicMock]:
        mock_config: MagicMock = MagicMock(spec=Config)
        mock_config.debug = MagicMock()
        mock_config.debug.production = False
        mock_config.deployed = False

        with patch(
            "acb.debug.depends.inject",
            lambda f: lambda *args, **kwargs: f(*args, config=mock_config, **kwargs),
        ):
            yield mock_config

    def test_init_debug_not_production(self, mock_config: MagicMock) -> None:
        with patch.object(debug, "configureOutput") as mock_configure:
            init_debug()

            mock_configure.assert_called_once()
            assert mock_configure.call_args[1]["prefix"] == "    debug:  "
            assert mock_configure.call_args[1]["includeContext"] is True
            assert "outputFunction" in mock_configure.call_args[1]
            assert "argToStringFunction" in mock_configure.call_args[1]


class TestDebugGlobal:
    def test_debug_global(self) -> None:
        from icecream import ic

        assert debug is ic

    def test_debug_with_different_log_levels(self) -> None:
        test_messages = [
            "Test info message",
            "Test warning message",
            "Test error message",
        ]

        with patch.object(debug, "outputFunction") as mock_output_func:
            for message in test_messages:
                debug(message)
            assert mock_output_func.call_count == len(test_messages)

    def test_get_calling_module_from_debug(self) -> None:
        from acb.debug import get_calling_module

        mock_frame = MagicMock()
        mock_frame.f_back.f_back.f_back.f_code.co_filename = "/path/to/test_module.py"

        with patch("logging.currentframe", return_value=mock_frame):
            result = get_calling_module()
            assert result is None or isinstance(result, Path)



"""Tests for the niquests adapter."""

import typing as t
from unittest.mock import AsyncMock, MagicMock

import pytest
from pydantic import SecretStr
from acb.adapters.requests.niquests import Requests, RequestsSettings


class MockNiquestsResponse:
    def __init__(
        self, status_code: int = 200, data: t.Any = None, json_data: t.Any = None
    ) -> None:
        self.status_code = status_code
        self._data = data
        self._json = json_data

    def raise_for_status(self) -> None:
        if self.status_code >= 400:
            raise Exception(f"HTTP Error: {self.status_code}")

    async def data(self) -> t.Any:
        return self._data

    async def json(self) -> t.Any:
        return self._json


class MockNiquestsAsyncSession:
    def __init__(self) -> None:
        self.base_url = ""
        self.get = AsyncMock()
        self.post = AsyncMock()
        self.put = AsyncMock()
        self.delete = AsyncMock()
        self.patch = AsyncMock()
        self.head = AsyncMock()
        self.options = AsyncMock()
        self.request = AsyncMock()
        self.close = AsyncMock()

        self.get.return_value = MockNiquestsResponse()
        self.post.return_value = MockNiquestsResponse()
        self.put.return_value = MockNiquestsResponse()
        self.delete.return_value = MockNiquestsResponse()
        self.patch.return_value = MockNiquestsResponse()
        self.head.return_value = MockNiquestsResponse()
        self.options.return_value = MockNiquestsResponse()
        self.request.return_value = MockNiquestsResponse()


@pytest.fixture
def mock_session() -> MockNiquestsAsyncSession:
    return MockNiquestsAsyncSession()


@pytest.fixture
def requests_adapter(mock_session: MockNiquestsAsyncSession) -> Requests:
    mock_config = MagicMock()
    mock_config.requests = RequestsSettings(
        base_url="https://api.example.com",
        timeout=30,
        auth=("username", SecretStr("password")),
    )

    mock_logger = MagicMock()

    adapter = Requests()
    adapter.config = mock_config
    adapter.logger = mock_logger

    object.__setattr__(adapter, "_client", mock_session)

    adapter.__class__.client = property(
        lambda self: object.__getattribute__(self, "_client")
    )

    mock_session.base_url = "https://api.example.com"

    return adapter


class TestRequestsSettings:
    def test_default_values(self) -> None:
        settings = RequestsSettings()
        assert settings.base_url == ""
        assert settings.timeout == 10
        assert settings.auth is None
        assert settings.cache_ttl == 7200

    def test_custom_values(self) -> None:
        from pydantic import SecretStr

        settings = RequestsSettings(
            base_url="https://api.example.com",
            timeout=30,
            auth=("username", SecretStr("password")),
            cache_ttl=3600,
        )

        assert settings.base_url == "https://api.example.com"
        assert settings.timeout == 30
        assert settings.auth is not None
        assert settings.auth[0] == "username"
        assert settings.auth[1].get_secret_value() == "password"
        assert settings.cache_ttl == 3600

        settings_no_auth = RequestsSettings(
            base_url="https://api.example.com",
            timeout=30,
            auth=None,
            cache_ttl=3600,
        )

        assert settings_no_auth.auth is None


@pytest.mark.asyncio
async def test_init(requests_adapter: Requests) -> None:
    # Create mock logger
    mock_logger = MagicMock()
    requests_adapter.logger = mock_logger

    await requests_adapter.init()
    mock_logger.debug.assert_called_once_with("Niquests adapter initialized")


@pytest.mark.asyncio
async def test_client_property(mock_session: MockNiquestsAsyncSession) -> None:
    adapter = Requests()
    adapter.config = MagicMock()
    adapter.config.requests.base_url = "https://api.example.com"
    object.__setattr__(adapter, "_client", mock_session)
    adapter.__class__.client = property(
        lambda self: object.__getattribute__(self, "_client")
    )

    mock_session.base_url = adapter.config.requests.base_url

    assert adapter.client.base_url == "https://api.example.com"

    mock_response = MockNiquestsResponse()
    adapter.client.get.return_value = mock_response

    result = await adapter.get("/test")
    adapter.client.get.assert_called_once()
    assert result == mock_response


@pytest.mark.asyncio
async def test_get(requests_adapter: Requests) -> None:
    mock_response = MockNiquestsResponse(json_data={"key": "value"})
    requests_adapter.client.get.return_value = mock_response  # type: ignore[attr-defined]

    response = await requests_adapter.get(
        url="/test",
        timeout=5,
        params={"param": "value"},
        headers={"header": "value"},
        cookies={"cookie": "value"},
    )

    requests_adapter.client.get.assert_called_once_with(  # type: ignore[attr-defined]
        "/test",
        timeout=5,
        params={"param": "value"},
        headers={"header": "value"},
        cookies={"cookie": "value"},
    )

    assert response == mock_response


@pytest.mark.asyncio
async def test_post(requests_adapter: Requests) -> None:
    mock_response = MockNiquestsResponse(status_code=201)
    requests_adapter.client.post.return_value = mock_response  # type: ignore[attr-defined]

    response = await requests_adapter.post(
        url="/test", data={"key": "value"}, timeout=5
    )

    requests_adapter.client.post.assert_called_once_with(  # type: ignore[attr-defined]
        "/test", data={"key": "value"}, json=None, timeout=5
    )

    assert response == mock_response


@pytest.mark.asyncio
async def test_put(requests_adapter: Requests) -> None:
    mock_response = MockNiquestsResponse()
    requests_adapter.client.put.return_value = mock_response  # type: ignore[attr-defined]

    response = await requests_adapter.put(url="/test", json={"key": "value"}, timeout=5)

    requests_adapter.client.put.assert_called_once_with(  # type: ignore[attr-defined]
        "/test", data=None, json={"key": "value"}, timeout=5
    )

    assert response == mock_response


@pytest.mark.asyncio
async def test_delete(requests_adapter: Requests) -> None:
    mock_response = MockNiquestsResponse(status_code=204)
    requests_adapter.client.delete.return_value = mock_response  # type: ignore[attr-defined]

    response = await requests_adapter.delete(url="/test", timeout=5)

    requests_adapter.client.delete.assert_called_once_with("/test", timeout=5)  # type: ignore[attr-defined]

    assert response == mock_response


@pytest.mark.asyncio
async def test_patch(requests_adapter: Requests) -> None:
    mock_response = MockNiquestsResponse()
    requests_adapter.client.patch.return_value = mock_response  # type: ignore[attr-defined]

    response = await requests_adapter.patch(
        url="/test", json={"key": "value"}, timeout=5
    )

    requests_adapter.client.patch.assert_called_once_with(  # type: ignore[attr-defined]
        "/test", timeout=5, data=None, json={"key": "value"}
    )

    assert response == mock_response


@pytest.mark.asyncio
async def test_head(requests_adapter: Requests) -> None:
    mock_response = MockNiquestsResponse()
    requests_adapter.client.head.return_value = mock_response  # type: ignore[attr-defined]

    response = await requests_adapter.head(url="/test", timeout=5)

    requests_adapter.client.head.assert_called_once_with("/test", timeout=5)  # type: ignore[attr-defined]

    assert response == mock_response


@pytest.mark.asyncio
async def test_options(requests_adapter: Requests) -> None:
    mock_response = MockNiquestsResponse()
    requests_adapter.client.options.return_value = mock_response  # type: ignore[attr-defined]

    response = await requests_adapter.options(url="/test", timeout=5)

    requests_adapter.client.options.assert_called_once_with("/test", timeout=5)  # type: ignore[attr-defined]

    assert response == mock_response


@pytest.mark.asyncio
async def test_request(requests_adapter: Requests) -> None:
    mock_response = MockNiquestsResponse()
    requests_adapter.client.request.return_value = mock_response  # type: ignore[attr-defined]

    response = await requests_adapter.request(
        method="GET", url="/test", json={"key": "value"}, timeout=5
    )

    requests_adapter.client.request.assert_called_once_with(  # type: ignore[attr-defined]
        "GET", "/test", data=None, json={"key": "value"}, timeout=5
    )

    assert response == mock_response


@pytest.mark.asyncio
async def test_close(requests_adapter: Requests) -> None:
    await requests_adapter.close()
    requests_adapter.client.close.assert_called_once()  # type: ignore[attr-defined]


@pytest.mark.asyncio
async def test_raise_for_status(requests_adapter: Requests) -> None:
    mock_response = MockNiquestsResponse(status_code=404)
    requests_adapter.client.get.return_value = mock_response  # type: ignore[attr-defined]

    with pytest.raises(Exception):
        await requests_adapter.get(url="/test")
